class ProfileConsole:
    def __init__(self, console):
        """
        The ProfileConsole class handles user profile related operations in the application.

        :param console: An instance of the Console class for inter-component communication.
        :type console: Console
        """
        self.console = console

    def edit_profile(self):
        """
        Handles editing of user profile.
        """
        if not self.console.bot.community_id:
            return self.console.on_error("You must select a community first.")
        
        self.console.print("Edit Profile")
        self.console.print("""
    1. Change Username
    2. Change Bio
    3. Change Profile Picture
    4. Change Background Picture
    5. Back
    """)
        choice = self.console.input(">>> ")

        try:
            if choice == "5":
                return self.console.menu.display()

            field_name = None
            new_value = None

            if choice == "1":
                field_name = "nickname"
                new_value = self.console.input("Enter new username: ")
            elif choice == "2":
                field_name = "content"
                new_value = self.console.input("Enter new bio: ")
            elif choice == "3":
                field_name = "icon"
                new_value = self.console.input("Enter new profile picture URL: ")
            elif choice == "4":
                field_name = "backgroundImage"
                new_value = self.console.input("Enter new background picture URL: ")
            else:
                self.console.print("Invalid option. Please try again.")
                return self.edit_profile()

            self.console.bot.community.edit_profile(**{field_name: new_value})
            self.console.print(f"{field_name.capitalize()} changed successfully to {new_value}.")
            self.console.sleep(2)

        except Exception as e:
            self.console.print(f"Error: {e}")
            self.console.sleep(2)

        return self.console.menu.display()



import textwrap

from os import system
from time import sleep
from .utilities.menu import Menu
from .entities.handlers import is_android, is_repl
from .utilities.community_console import CommunityConsole
from .utilities.profile_console import ProfileConsole
from .utilities.chat_console import ChatConsole
from colorama import Fore, Style

class Console:
    def __init__(self, bot):
        """
        The Console class is responsible for the overall operation of the terminal application.

        :param bot: An instance of the bot which will interact with the Amino API.
        :type bot: Bot
        """
        self.bot = bot
        self.indent_size = self.fetch_indent_size()
        self.menu = Menu(self)
        self.community_console = CommunityConsole(self)
        self.profile_console = ProfileConsole(self)
        self.chat_console = ChatConsole(self)

    
    def fetch_indent_size(self):
        """
        Checks if the application is running on an Android device or a REPL environment and sets the appropriate
        indentation size.
        """
        return 0 if is_android() or is_repl() else 20

    def print(self, text=""):
        """
        Prints a string with a specified indentation.

        :param text: The string to be printed.
        :type text: str
        :param amount: The number of spaces to indent.
        :type amount: int
        """
        print(textwrap.indent(text, ' ' * self.indent_size))

    def error_print(self, text=""):
        """
        Prints a string with a specified indentation and a red color.

        :param text: The string to be printed.
        :type text: str
        """
        print(Fore.RED + textwrap.indent(text, ' ' * self.indent_size) + Style.RESET_ALL)

    def on_error(self, error):
        """
        Displays an error message to the user and prompts them to press enter to return to the main menu.
        """
        self.error_print(error)
        input(f"{' ' * self.indent_size}Press enter to return to the main menu.")
        self.fetch_menu()

    def sleep(self, seconds):
        """
        Sleeps for a specified number of seconds.

        :param seconds: The number of seconds to sleep.
        :type seconds: int
        """
        sleep(seconds)

    def clear(self):
        """
        Clears the console screen irrespective of the platform (Windows/Linux).
        """
        clear_command = "cls || clear"
        system(clear_command)

    def input(self, text):
        """
        Accepts user input prefixed with a fixed number of spaces for consistent look and feel.

        :param text: The prompt to be displayed to the user.
        :type text: str
        :return: User input as a string.
        """
        return input(" " * self.indent_size + text)

    def fetch_menu(self):
        """
        Displays the main menu of the application to the user and processes their input.
        """
        self.menu.display()



from typing import Union

"""
class LinkInfo:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data               = data
        self.linkInfoV2         = {}
        self.path               = None
        self.extensions         = {}
        self.objectId           = None
        self.shareURLShortCode  = None
        self.targetCode         = None
        self.ndcId              = None
        self.comId              = None
        self.fullPath           = None
        self.shortCode          = None
        self.objectType         = None

        if isinstance(data, dict):
            self.linkInfoV2:        dict = self.data.get("linkInfoV2", self.linkInfoV2)
            self.path:              Union[str, None] = self.data.get("path", self.path)             or self.linkInfoV2.get("path", self.path)
            self.extensions:        dict = self.data.get("extensions", self.extensions)             or self.linkInfoV2.get("extensions", self.extensions)
            self.objectId:          Union[str, None] = self.data.get("objectId", self.objectId)     or self.extensions.get("linkInfo", {}).get("objectId", self.objectId)
            self.shareURLShortCode: Union[str, None] = self.data.get("shareURLShortCode", self.shareURLShortCode) or self.extensions.get("linkInfo", {}).get("shareURLShortCode", self.shareURLShortCode)
            self.targetCode:        Union[str, None] = self.data.get("targetCode", self.targetCode) or self.extensions.get("linkInfo", {}).get("targetCode", self.targetCode)
            self.ndcId:             Union[int, None] = self.data.get("ndcId", self.ndcId)           or self.extensions.get("linkInfo", {}).get("ndcId", self.ndcId)
            self.comId:             Union[int, None] = self.ndcId
            self.fullPath:          Union[str, None] = self.data.get("fullPath", self.fullPath)     or self.extensions.get("linkInfo", {}).get("fullPath", self.fullPath)
            self.shortCode:         Union[str, None] = self.data.get("shortCode", self.shortCode)   or self.extensions.get("linkInfo", {}).get("shortCode", self.shortCode)
            self.objectType:        Union[str, None] = self.data.get("objectType", self.objectType) or self.extensions.get("linkInfo", {}).get("objectType", self.objectType)

    def json(self) -> Union[dict, str]:
        return self.data
"""
class LinkInfo:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    
    @property
    def linkInfoV2(self) -> Union[str, None]:
        """Returns the linkInfoV2 of the API response."""
        return self.data.get("linkInfoV2")
    
    @property
    def path(self) -> Union[int, None]:
        """Returns the path of the API response."""
        return self.data.get("path") or self.linkInfoV2.get("path")
    
    @property
    def extensions(self) -> Union[dict, None]:
        """Returns the extensions of the API response."""
        return self.data.get("extensions") or self.linkInfoV2.get("extensions")
    
    @property
    def objectId(self) -> Union[str, None]:
        """Returns the objectId of the API response."""
        return self.data.get("objectId") or self.extensions.get("linkInfo", {}).get("objectId")
    
    @property
    def shareURLShortCode(self) -> Union[str, None]:
        """Returns the shareURLShortCode of the API response."""
        return self.data.get("shareURLShortCode") or self.extensions.get("linkInfo", {}).get("shareURLShortCode")
    
    @property
    def targetCode(self) -> Union[str, None]:
        """Returns the targetCode of the API response."""
        return self.data.get("targetCode") or self.extensions.get("linkInfo", {}).get("targetCode")
    
    @property
    def ndcId(self) -> Union[int, None]:
        """Returns the ndcId of the API response."""
        return self.data.get("ndcId") or self.extensions.get("linkInfo", {}).get("ndcId")
    
    @property
    def comId(self) -> Union[int, None]:
        """Returns the comId of the API response."""
        return self.ndcId
    
    @property
    def fullPath(self) -> Union[str, None]:
        """Returns the fullPath of the API response."""
        return self.data.get("fullPath") or self.extensions.get("linkInfo", {}).get("fullPath")
    
    @property
    def shortCode(self) -> Union[str, None]:
        """Returns the shortCode of the API response."""
        return self.data.get("shortCode") or self.extensions.get("linkInfo", {}).get("shortCode")
    
    @property
    def objectType(self) -> Union[str, None]:
        """Returns the objectType of the API response."""
        return self.data.get("objectType") or self.extensions.get("linkInfo", {}).get("objectType")
    
    def json(self) -> Union[dict, str]:
        """Returns the JSON data of the API response."""
        return self.data
    
    def __repr__(self) -> str:
        """Returns the representation of the Link Info response."""
        return f"<LinkInfo data={self.data}>"


from typing import List


class ChatThreadExtensions:
    def __init__(self, data: dict):
        try:
            self.data = data.get("extensions", data)
        except AttributeError:
            self.data = None

    
    def _check_extensions(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper
    

    @property
    @_check_extensions
    def is_view_only(self) -> bool:
        """Returns whether the chat is view only"""
        return self.data.get("viewOnly")

    @property
    @_check_extensions
    def coHosts(self) -> List[str]:
        """Returns a list of co-host uids"""
        return self.data.get("coHost")
    
    @property
    @_check_extensions
    def coHost(self) -> str:
        #NOTE: This will be removed in the future
        """Returns the co-host uid"""
        return self.data.get("coHost")

    @property
    @_check_extensions
    def language(self) -> str:
        """Returns the language of the chat"""
        return self.data.get("language")
    
    @property
    @_check_extensions
    def is_members_can_invite(self) -> bool:
        """Returns whether members can invite others"""
        return self.data.get("membersCanInvite")
    
    @property
    @_check_extensions
    def screening_room_permission(self) -> dict:
        """Returns the screening room permission"""
        return self.data.get("screeningRoomPermission")
    
    @property
    @_check_extensions
    def background(self) -> str:
        """Returns the background of the chat"""
        try:
            return self.data.get("bm")[1]
        except IndexError:
            return None

    @property
    @_check_extensions
    def avchat_member_user_id_list(self) -> List[str]:
        """Returns a list of avchat member uids"""
        return self.data.get("avchatMemberUidList")

    @property
    @_check_extensions
    def screening_room_host_user_id(self) -> str:
        """Returns the screening room host uid"""
        return self.data.get("screeningRoomHostUid")

    @property
    @_check_extensions
    def visibility(self) -> int:
        """Returns the visibility of the chat"""
        return self.data.get("visibility")

    @property
    @_check_extensions
    def banned_user_ids(self) -> List[str]:
        """Returns a list of banned member uids"""
        return self.data.get("bannedMemberUidList")

    @property
    @_check_extensions
    def last_message_time(self) -> int:
        """Returns the last members summary update time"""
        return self.data.get("lastMembersSummaryUpdateTime")

    @property
    @_check_extensions
    def is_fans_only(self) -> bool:
        """Returns whether the chat is fans only"""
        return self.data.get("fansOnly")

    @property
    @_check_extensions
    def announcement(self) -> str:
        """Returns the announcement of the chat"""
        return self.data.get("announcement")

    @property
    @_check_extensions
    def last_update_time(self) -> int:
        """Returns the channel type last created time"""
        return self.data.get("channelTypeLastCreatedTime")
    
    @property
    @_check_extensions
    def avchatId(self) -> str:
        """Returns the avchat id"""
        return self.data.get("avchatId")
    
    @property
    @_check_extensions
    def channel_type(self) -> int:
        """Returns the channel type"""
        return self.data.get("channelType")
    
    @property
    @_check_extensions
    def is_announcement_pinned(self) -> bool:
        """Returns whether the chat announcement is pinned"""
        return self.data.get("pinAnnouncement")
    
    @property
    @_check_extensions
    def voice_chat_join_type(self) -> int:
        """Returns the vv chat join type"""
        return self.data.get("vvChatJoinType")
    
    @property
    @_check_extensions
    def tipping_perm_status(self) -> int:
        """Returns the tipping permission status"""
        return self.data.get("tippingPermStatus")
    
    @_check_extensions
    def json(self) -> dict:
        """Returns the json data"""
        return self.data


class ChatThreadExtensionsList:
    def __init__(self, data: list):
        try:
            self.data = data
        except AttributeError:
            self.data = None

    @property
    def is_view_only(self) -> List[bool]:
        """Returns a list of whether the chat is view only"""
        return [i.get("viewOnly") if i else None for i in self.data]
    
    @property
    def co_host_user_ids(self) -> List[List[str]]:
        """Returns a list of co-host uids"""
        return [i.get("coHost") if i else None for i in self.data]
    
    @property
    def language(self) -> List[str]:
        """Returns a list of the language of the chat"""
        return [i.get("language") if i else None for i in self.data]
    
    @property
    def is_members_can_invite(self) -> List[bool]:
        """Returns a list of whether members can invite others"""
        return [i.get("membersCanInvite") if i else None for i in self.data]
    
    @property
    def screening_room_permission(self) -> List[dict]:
        """Returns a list of the screening room permission"""
        return [i.get("screeningRoomPermission") if i else None for i in self.data]
    
    @property
    def background(self) -> List[str]:
        """Returns a list of the background of the chat"""
        return [i.get("bm")[1] if i and isinstance(i.get("bm"), list) else None for i in self.data]
    
    @property
    def avchat_member_user_id_list(self) -> List[List[str]]:
        """Returns a list of avchat member uids"""
        return [i.get("avchatMemberUidList") if i else None for i in self.data]
    
    @property
    def screening_room_host_user_id(self) -> List[str]:
        """Returns a list of the screening room host uid"""
        return [i.get("screeningRoomHostUid") if i else None for i in self.data]
    
    @property
    def visibility(self) -> List[int]:
        """Returns a list of the visibility of the chat"""
        return [i.get("visibility") if i else None for i in self.data]
    
    @property
    def banned_user_ids(self) -> List[List[str]]:
        """Returns a list of banned member uids"""
        return [i.get("bannedMemberUidList") if i else None for i in self.data]
    
    @property
    def last_message_time(self) -> List[int]:
        """Returns a list of the last members summary update time"""
        return [i.get("lastMembersSummaryUpdateTime") if i else None for i in self.data]
    
    @property
    def is_fans_only(self) -> List[bool]:
        """Returns a list of whether the chat is fans only"""
        return [i.get("fansOnly") if i else None for i in self.data]
    
    @property
    def announcement(self) -> List[str]:
        """Returns the announcement of the chat"""
        return [i.get("announcement") if i else None for i in self.data]

    @property
    def last_update_time(self) -> List[int]:
        """Returns the channel type last created time"""
        return [i.get("channelTypeLastCreatedTime") if i else None for i in self.data]
    
    @property
    def avchatId(self) -> List[str]:
        """Returns the avchat id"""
        return [i.get("avchatId") if i else None for i in self.data]
    
    @property
    def channel_type(self) -> List[int]:
        """Returns the channel type"""
        return [i.get("channelType") if i else None for i in self.data]
    
    @property
    def is_announcement_pinned(self) -> List[bool]:
        """Returns whether the chat announcement is pinned"""
        return [i.get("pinAnnouncement") if i else None for i in self.data]
    
    @property
    def voice_chat_join_type(self) -> List[int]:
        """Returns the vv chat join type"""
        return [i.get("vvChatJoinType") if i else None for i in self.data]
    
    @property
    def tipping_perm_status(self) -> List[int]:
        """Returns the tipping permission status"""
        return [i.get("tippingPermStatus") if i else None for i in self.data]
    
    def json(self) -> dict:
        """Returns the json data"""
        return self.data


class MemberSummary:
    def __init__(self, data: dict):
        try:
            self.data = data.get("membersSummary", data)
        except AttributeError:
            self.data = None


    def _check_memberSummary(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper


    @property
    @_check_memberSummary
    def status(self) -> List[int]:
        """Returns a list of member statuses"""
        return [i.get("status") for i in self.data]

    @property
    @_check_memberSummary
    def uid(self) -> List[str]:
        """Returns a list of member uids"""
        return [i.get("uid") for i in self.data]
    
    @property
    @_check_memberSummary
    def userId(self) -> List[str]:
        """Returns a list of member user ids"""
        return self.uid

    @property
    @_check_memberSummary
    def membership_status(self) -> List[int]:
        """Returns a list of member membership statuses"""
        return [i.get("membershipStatus") for i in self.data]

    @property
    @_check_memberSummary
    def role(self) -> List[int]:
        """Returns a list of member roles"""
        return [i.get("role") for i in self.data]

    @property
    @_check_memberSummary
    def nickname(self) -> List[str]:
        """Returns a list of member nicknames"""
        return [i.get("nickname") for i in self.data]
    
    @property
    @_check_memberSummary
    def username(self) -> List[str]:
        """Returns a list of member usernames"""
        return self.nickname

    @property
    @_check_memberSummary
    def icon(self) -> List[str]:
        """Returns a list of member icons"""
        return [i.get("icon") for i in self.data]
    
    @property
    @_check_memberSummary
    def avatar(self) -> List[str]:
        """Returns a list of member avatars"""
        return self.icon

    @_check_memberSummary
    def json(self) -> List[dict]:
        """Returns a list of member summaries"""
        return self.data


class MemberSummaryList:
    def __init__(self, data: list):
        try:
            self.data = data
        except AttributeError:
            self.data = None
    
    @property
    def status(self) -> List[List[int]]:
        """Returns a list of member statuses"""
        return [[x.get("status") for x in y] for y in self.data]
    
    @property
    def uid(self) -> List[str]:
        """Returns a list of member uids"""
        return [[x.get("uid") for x in y] for y in self.data]
    
    @property
    def userId(self) -> List[str]:
        """Returns a list of member user ids"""
        return self.uid
    
    @property
    def membershipStatus(self) -> List[int]:
        """Returns a list of member membership statuses"""
        return [[x.get("membershipStatus") for x in y] for y in self.data]
    
    @property
    def role(self) -> List[int]:
        """Returns a list of member roles"""
        return [[x.get("role") for x in y] for y in self.data]
    
    @property
    def nickname(self) -> List[str]:
        """Returns a list of member nicknames"""
        return [[x.get("nickname") for x in y] for y in self.data]
    
    @property
    def username(self) -> List[str]:
        """Returns a list of member usernames"""
        return self.nickname
    
    @property
    def icon(self) -> List[str]:
        """Returns a list of member icons"""
        return [[x.get("icon") for x in y] for y in self.data]
    
    @property
    def avatar(self) -> List[str]:
        """Returns a list of member avatars"""
        return self.icon
    
    def json(self) -> List[dict]:
        """Returns a list of member summaries"""
        return self.data


class ChatThread: 
    def __init__(self, data: dict):
        try:
            self.data = data.get("thread", data)
        except AttributeError:
            self.data = None

    
    @property
    def user_added_topic_list(self) -> List[str]:
        """Returns a list of user added topics"""
        return self.data.get("userAddedTopicList")
    
    @property
    def uid(self) -> str:
        """Returns the host user id"""
        return self.data.get("uid")
    
    @property
    def host_user_id(self) -> str:
        """Returns the host user id"""
        return self.uid
    
    @property
    def hostUserId(self) -> str:
        #NOTE: This will be removed in the future
        return self.uid
    
    @property
    def members_quota(self) -> int:
        """Returns the members quota"""
        return self.data.get("membersQuota")
    
    @property
    def members_summary(self) -> MemberSummary:
        """Returns the members summary"""
        return MemberSummary(self.data.get("membersSummary"))
    
    @property
    def threadId(self) -> str:
        """Returns the thread id"""
        return self.data.get("threadId")
    
    @property
    def chatId(self) -> str:
        """Returns the thread id"""
        return self.threadId
    
    @property
    def keywords(self) -> List[str]:
        """Returns a list of keywords"""
        return self.data.get("keywords")
    
    @property
    def members_count(self) -> int:
        """Returns the members count"""
        return self.data.get("membersCount")
    
    @property
    def strategy_info(self) -> dict:
        """Returns the strategy info"""
        return self.data.get("strategyInfo")
    
    @property
    def is_pinned(self) -> bool:
        """Returns whether the chat is pinned"""
        return self.data.get("isPinned")
    
    @property
    def title(self) -> str:
        """Returns the chat title"""
        return self.data.get("title")
    
    @property
    def membership_status(self) -> str:
        """Returns the membership status"""
        return self.data.get("membershipStatus")
    
    @property
    def content(self) -> str:
        """Returns the chat content"""
        return self.data.get("content")
    
    @property
    def is_hidden_required(self) -> bool:
        """Returns whether the chat needs to be hidden"""
        return self.data.get("needHidden")

    @property
    def alert_option(self) -> int:
        """Returns the alert option"""
        return self.data.get("alertOption")
    
    @property
    def last_read_time(self) -> str:
        """Returns the last read time"""
        return self.data.get("lastReadTime")
    
    @property
    def type(self) -> int:
        """Returns the chat type"""
        return self.data.get("type")
    
    @property
    def status(self) -> int:
        """Returns the chat status"""
        return self.data.get("status")
    
    @property
    def is_published_to_global(self) -> bool:
        """Returns whether the chat is published to global"""
        return self.data.get("publishToGlobal")
    
    @property
    def modified_time(self) -> str:
        """Returns the modified time"""
        return self.data.get("modifiedTime")
    
    @property
    def last_message_summary(self) -> dict:
        """Returns the last message summary"""
        return self.data.get("lastMessageSummary")
    
    @property
    def extensions(self) -> ChatThreadExtensions:
        """Returns the chat thread extensions"""
        return ChatThreadExtensions(self.data.get("extensions"))

    def json(self) -> dict:
        """Returns the chat thread json"""
        return self.data


class ChatThreadList:
    def __init__(self, data: dict):
        self.data = data.get("threadList", data)

    def parser(self) -> List[ChatThread]:
        """Returns a list of ChatThread objects"""
        return [ChatThread(x) for x in self.data]
    
    
    @property
    def extensions(self) -> ChatThreadExtensions:
        """Returns the chat thread extensions"""
        return ChatThreadExtensionsList([x.get("extensions") for x in self.data])
    
    @property
    def members_summary(self) -> MemberSummary:
        """Returns the members summary"""
        return MemberSummaryList([x.get("membersSummary") for x in self.data])
    
    @property
    def user_added_topic_list(self) -> List[str]:
        """Returns a list of user added topics"""
        return [x.user_added_topic_list for x in self.parser()]
    
    @property
    def uid(self) -> List[str]:
        """Returns a list of host user ids"""
        return [x.uid for x in self.parser()]
    
    @property
    def host_user_id(self) -> List[str]:
        """Returns a list of host user ids"""
        return self.uid
    
    @property
    def members_quota(self) -> List[int]:
        """Returns a list of members quotas"""
        return [x.members_quota for x in self.parser()]
    
    @property
    def threadId(self) -> List[str]:
        """Returns a list of thread ids"""
        return [x.threadId for x in self.parser()]
    
    @property
    def chatId(self) -> List[str]:
        """Returns a list of thread ids"""
        return self.threadId
    
    @property
    def keywords(self) -> List[List[str]]:
        """Returns a list of keywords"""
        return [x.keywords for x in self.parser()]
    
    @property
    def members_count(self) -> List[int]:
        """Returns a list of members counts"""
        return [x.members_count for x in self.parser()]
    
    @property
    def strategy_info(self) -> List[dict]:
        """Returns a list of strategy info"""
        return [x.strategy_info for x in self.parser()]
    
    @property
    def is_pinned(self) -> List[bool]:
        """Returns a list of whether the chat is pinned"""
        return [x.is_pinned for x in self.parser()]
    
    @property
    def title(self) -> List[str]:
        """Returns a list of chat titles"""
        return [x.title for x in self.parser()]
    
    @property
    def membership_status(self) -> List[str]:
        """Returns a list of membership statuses"""
        return [x.membership_status for x in self.parser()]
    
    @property
    def content(self) -> List[str]:
        """Returns a list of chat contents"""
        return [x.content for x in self.parser()]
    
    @property
    def is_hidden_required(self) -> List[bool]:
        """Returns a list of whether the chat needs to be hidden"""
        return [x.is_hidden_required for x in self.parser()]
    
    @property
    def alert_option(self) -> List[int]:
        """Returns a list of alert options"""
        return [x.alert_option for x in self.parser()]
    
    @property
    def last_read_time(self) -> List[str]:
        """Returns a list of last read times"""
        return [x.last_read_time for x in self.parser()]
    
    @property
    def type(self) -> List[int]:
        """Returns a list of chat types"""
        return [x.type for x in self.parser()]
    
    @property
    def status(self) -> List[int]:
        """Returns a list of chat statuses"""
        return [x.status for x in self.parser()]
    
    @property
    def is_published_to_global(self) -> List[bool]:
        """Returns a list of whether the chat is published to global"""
        return [x.is_published_to_global for x in self.parser()]
    
    @property
    def modified_time(self) -> List[str]:
        """Returns a list of modified times"""
        return [x.modified_time for x in self.parser()]
    
    @property
    def last_message_summary(self) -> List[dict]:
        """Returns a list of last message summaries"""
        return [x.last_message_summary for x in self.parser()]
    
    def json(self) -> dict:
        """Returns the chat thread list json"""
        return self.data
    



from os import path
from threading import Thread
from typing import Optional, Union
from time import perf_counter, time
from logging.handlers import RotatingFileHandler
from logging import Logger, getLogger, Formatter, DEBUG
import requests

from .ext.console import *
from .ext.entities import *
from .ext.account import Account
from .ext.socket import WSClient
from .ext.community import Community
from .ext.global_client import Global
from .ext.utilities.generate import Generator
from .ext.utilities.request_handler import RequestHandler

__all__ = (
    'Bot',
)


class Bot(WSClient, Global):
    """
    Bot class that interacts with aminoapps API.

    This class extends `WSClient` and `Global` classes, allowing the bot to use WebSocket functionality and global client features.

    Special Attributes:
    __slots__ : tuple
        A tuple containing a fixed set of attributes to optimize memory usage.

    Attributes:
    _debug : bool
        Whether or not debug mode is enabled.
    _console_enabled : bool
        Whether or not the CONSOLE is enabled.
    _cooldown_message : str
        The default cooldown message used when a command is on cooldown.
    _intents : bool
        Whether or not intents are enabled.
    _is_ready : bool
        Whether the bot is ready after successful login.
    _userId : str
        The ID of the user associated with the bot.
    _sid : str
        The session ID of the client.
    _cached : bool
        Whether the login credentials are cached.
    logger : Logger
        The logger object.
    cache : Cache
        An instance of the Cache class for caching data.
    command_prefix : Optional[str]
        The prefix used for bot commands.
    community_id : Union[str, int]
        The ID of the community associated with the bot.
    generate : Generator
        An instance of the Generator class for generating data.
    online_status : bool
        Whether the bot's online status is enabled.
    device_id : Optional[str]
        The device ID used for logging in. If not provided, it will be generated using the Generator class.
    _is_authenticated : bool
        Whether the bot is authenticated.
    request : RequestHandler
        An instance of the RequestHandler class for handling API requests.
    community : Community
        An instance of the Community class for community-related actions.
    account : Account
        An instance of the Account class for account-related actions.
    profile : UserProfile
        An instance of the UserProfile class representing the bot's user profile.
    """
    __slots__ = (
        '_debug',
        '_console_enabled',
        '_cooldown_message',
        '_intents',
        '_is_ready',
        '_userId',
        '_sid',
        '_secret',
        '_cached',
        'cache',
        'logger',
        'command_prefix',
        'community_id',
        'debug_log',
        'generate',
        'online_status',
        'device_id',
        '_is_authenticated'
        'request',
        'community',
        'account',
        'profile',
    )
    def __init__(
        self,
        command_prefix: Optional[str] = "!",
        community_id: Union[str, int] = None,
        console_enabled: bool = False,
        debug_log: bool = False,
        device_id: str = None,
        intents: bool = False,
        online_status: bool = False,
        proxy: str = None,
        hash_prefix: Union[str, int] = 52,
        device_key: str = "AE49550458D8E7C51D566916B04888BFB8B3CA7D",
        signature_key: str  = "EAB4F1B9E3340CD1631EDE3B587CC3EBEDF1AFA9",
        service_key: str = None,
        ) -> None:
        """
        `Bot` - This is the main client.

        `**Parameters**``
        - `command_prefix` - The prefix to use for commands. `Defaults` to `!`.
        - `community_id` - The community id to use for the bot. `Defaults` to `None`.
        - `console_enabled` - Whether to enable the console. `Defaults` to `True`.
        - `debug_log` - Whether to enable logging to file. `Defaults` to `False`.
        - `device_id` - The device id to use for the bot. `Defaults` to `None`.
        - `intents` - Avoids receiving events that you do not need. `Defaults` to `False`.
        - `online_status` - Whether to set the bot's online status to `online`. `Defaults` to `True`.
        - `proxy` - The proxy to use for the bot. `Defaults` to `None`.
        - `hash_prefix` - The hash prefix to use for the bot. `Defaults` to `19`.
        - `device_key` - The device key to use for the bot.
        - `signature_key` - The signature key to use for the bot.

        ----------------------------
        When should I use `Bot` instead of `Client`?

        - If you want to create a bot that responds to commands or events.
        - `Client` does not respond to commands or events.

        ----------------------------

        How do I login to my bot account?

        - You can login to your bot account by using the `run` method.

        ----------------------------

        How do I use the `run` method?
        
        - It's simple! Just use the `run` method like this:

            ```py
            bot = Bot()
            #NOTE: Keep in mind bot.run() should be the last line of your code.
            bot.run(email="email", password="password")
            ```

        ----------------------------
        How do I login to my bot account with a proxy?

        - It's as easy as adding a `proxy` parameter to the `bot` class.

            ```py
            bot = Bot(proxy="http://username:password@ip:port")

            bot.run(email="email", password="password")
            ```
        ----------------------------
        How do I know if my bot is ready?

        - You can check if your bot is ready by using the `is_ready` attribute.
        - `is_ready` is a `bool` that is `True` if the bot is ready and `False` if the bot is not ready.
        - Alternatively, you can use the `on_ready` event.
        - `on_ready` is an event that is called when the bot is ready.

        ----------------------------
        How do I use the `on_ready` event?

        - You can use the `ready` event like this:

            ```py
            bot = Bot()

            @bot.on("on_ready")
            def on_ready():
                print(f"Logged in as {bot.profile.username}(bot.profile.userId)")

            bot.run(email="email", password="password")
            ```
        ----------------------------
        How do I use the is_ready attribute?

        - You can use the `is_ready` attribute like this:

            ```py
            bot = Bot()

            bot.run(email="email", password="password")

            if bot.is_ready:
                print(f"Logged in as {bot.profile.username}(bot.profile.userId)")

            ```
        ----------------------------
        How do I set the community id for my bot?

        - If you know the community id, you can use the `set_community_id` method to set the community id for your bot.
        - Alternatively, if you do not know the community id, you can use the `fetch_community_id` method to get the community id from the community's link.

        ----------------------------
        How do I use the `set_community_id` method?

        - You can use the `set_community_id` method like this:

            ```py
            bot = Bot()

            bot.run(email="email", password="password")
            bot.set_community_id(community_id=123456789)
            ```
        ----------------------------
        How do I use the `fetch_community_id` method?

        - You can use the `fetch_community_id` method like this:

            ```py
            bot = Bot()

            bot.run(email="email", password="password")
            bot.fetch_community_id(community_link="https://aminoapps.com/c/OnePiece")
            ```
        ----------------------------
        Ok, I'm ready to make my first bot! What do I do now?
        - You can start by making a `command`.
        - They are functions that are called when a user sends a message that starts with the `command_prefix`.
        - You can make a command like this:

            ```py
            from pymino import Bot
            from pymino.ext import *

            bot = Bot()

            @bot.command(command_name="ping")
            def ping_command(ctx: Context):
                ctx.send("Pong!")

            bot.run(email="email", password="password")
            ```
        ----------------------------
        What is my command prefix?
        - Your command prefix by default is `!`.
        - You can change your command prefix by using the `command_prefix` parameter in the `Bot` class.
        - You can change your command prefix like this:

            ```py
            bot = Bot(command_prefix=".")
            # Now your command prefix is "." instead of "!".
            ```
        ----------------------------
        Can I use multiple names for my command?
        - Yes! You can use multiple names for your command by using the `aliases` parameter in the `command` decorator.
        - You can use multiple names for your command like this:

            ```py
            from pymino import Bot
            from pymino.ext import *

            bot = Bot()
            #NOTE: This will register the command as "ping", "p", and "pong".
            # Meaning the command will be called if the message starts with "!ping", "!p", or "!pong".
            @bot.command(command_name="ping", aliases=["p", "pong"])
            def ping_command(ctx: Context):
                ctx.send("Pong!")

            bot.run(email="email", password="password")
            ```
        ----------------------------
        How about cooldowns? Can I use them?
        - Yes! You can use cooldowns for your commands by using the `cooldown` parameter in the `command` decorator.
        - You can use cooldowns like this:

            ```py
            from pymino import Bot
            from pymino.ext import *

            bot = Bot()
            #NOTE: This will set the cooldown for the command to 5 seconds.
            # The cooldown is user based, meaning that each user will have their own cooldown.
            @bot.command(command_name="ping", cooldown=5)
            def ping_command(ctx: Context):
                ctx.send("Pong!")

            bot.run(email="email", password="password")
            ```
        ----------------------------
        Is there a help command built in?
        - Yes! There is a built in help command that you can use.
        - It will return a list of all the commands that the bot has and their descriptions.
        
        How do I set the description for my command?
        - You can set the description for your command by using the `command_description` parameter in the `command` decorator.
        - You can set the description for your command like this:

            ```py
            from pymino import Bot
            from pymino.ext import *

            bot = Bot()
            # This way when the user uses the help command, it will return the description for the command.
            @bot.command(command_name="ping", command_description="This command will return pong.")
            def ping_command(ctx: Context):
                ctx.send("Pong!")

            bot.run(email="email", password="password")
            ```
        """
        self.__local_cache__:       Cache = Cache(f"{path.dirname(path.realpath(__file__))}/cache")
        self.__device_key__:        str = self.__local_cache__.get("device_key", device_key)
        self.__key__:        str = self.__local_cache__.get("KEY", service_key)
        self.__signature_key__:     str = self.__local_cache__.get("signature_key", signature_key)
        if not self.__key__:
            raise MissingServiceKey
        if not all([self.__device_key__, self.__signature_key__]):
            raise MissingDeviceKeyOrSignatureKey

        self._debug:            bool = check_debugger()
        self._console_enabled:  bool = console_enabled
        self._cooldown_message: Optional[str] = None
        self._is_authenticated: bool = False
        self._intents:          bool = intents
        self._is_ready:         bool = False
        self._userId:           str = None
        self._sid:              str = None
        self._cached:           bool = False
        self.cache:             Cache = Cache("cache")
        self.command_prefix:    Optional[str] = command_prefix
        if self.command_prefix == "":
            raise InvalidCommandPrefix()

        self.logger:            Optional[Logger] = self._create_logger() if debug_log else None
        self.community_id:      Union[str, int] = community_id
        self.generate:          Generator = Generator(hash_prefix, self.__device_key__, self.__signature_key__, self.__key__)
        self.online_status:     bool = online_status
        self.device_id:         Optional[str] = device_id or self.generate.device_id()
        self.request:           RequestHandler = RequestHandler(
                                bot = self,
                                proxy=proxy,
                                generator=self.generate,
                                KEY=self.__key__
                                )
        self.community:         Community = Community(
                                bot = self,
                                session=self.request,
                                community_id=self.community_id
                                )
        self.account:           Account = Account(
                                session=self.request
                                )

        if self.community_id:   self.set_community_id(community_id)

        super().__init__()

    def __repr__(self):
        """
        Returns a string representation of the Bot object.

        :return: A string representation of the Bot object.
        :rtype: str
        """
        return f"Bot(command_prefix='{self.command_prefix}', community_id={self.community_id}, device_id='{self.device_id}')"

    def __str__(self):
        """
        Returns a user-friendly string representation of the Bot object.

        :return: A user-friendly string representation of the Bot object.
        :rtype: str
        """
        return f"Bot: Prefix='{self.command_prefix}', Community ID={self.community_id}, Device ID='{self.device_id}'"

    def __iter__(self) -> iter:
        """
        Allows iteration over the Bot object.

        :return: An iterator for the Bot object.
        :rtype: iter
        """
        return iter(self.__slots__)

    @property
    def debug(self) -> bool:
        """
        Whether or not debug mode is enabled.

        :return: True if debug mode is enabled, False otherwise.
        :rtype: bool

        This property returns whether or not debug mode is enabled. Debug mode can be used to enable additional logging and
        debug information during development.

        **Note:** This property only returns the debug mode state and cannot be used to set the debug mode state. To set the
        debug mode state, use the `self._debug` attribute directly.
        """
        return self._debug

    @debug.setter
    def debug(self, value: bool) -> None:
        """
        Sets the debug mode state.

        :param value: True to enable debug mode, False to disable it.
        :type value: bool
        :return: None

        This setter sets the debug mode state. Debug mode can be used to enable additional logging and debug information
        during development.

        **Note:** This setter only sets the debug mode state and cannot be used to retrieve the debug mode state. To retrieve
        the debug mode state, use the `self.debug` property.
        """
        self._debug = value

    @property
    def console_enabled(self) -> bool:
        """
        Whether or not the CONSOLE is enabled.

        :return: True if the CONSOLE is enabled, False otherwise.
        :rtype: bool

        This property returns whether or not the CONSOLE is enabled. The CONSOLE can be used to interact with the bot and access
        additional features such as the console.

        **Note:** This property only returns the CONSOLE state and cannot be used to set the CONSOLE state. To set the CONSOLE state, use
        the `self._console_enabled` attribute directly.
        """
        return self._console_enabled

    @console_enabled.setter
    def console_enabled(self, value: bool) -> None:
        """
        Sets the CONSOLE state.

        :param value: True to enable the CONSOLE, False to disable it.
        :type value: bool
        :return: None

        This setter sets the CONSOLE state. The CONSOLE can be used to interact with the bot and access additional features such as
        the console.

        **Note:** This setter only sets the CONSOLE state and cannot be used to retrieve the CONSOLE state. To retrieve the CONSOLE state,
        use the `self.console_enabled` property.
        """
        self._console_enabled = value

    @property
    def intents(self) -> bool:
        """
        Whether or not intents are enabled.

        :return: True if intents are enabled, False otherwise.
        :rtype: bool

        This property returns whether or not intents are enabled. Intents allow the bot to use additional features such as
        `ctx.wait_for_message()`.

        **Note:** This property only returns the intents state and cannot be used to set the intents state. To set the intents
        state, use the `self._intents` attribute directly.
        """
        return self._intents

    @intents.setter
    def intents(self, value: bool) -> None:
        """
        Sets the intents state.

        :param value: True to enable intents, False to disable them.
        :type value: bool
        :return: None

        This setter sets the intents state. Intents allow the bot to use additional features such as `ctx.wait_for_message()`.

        **Note:** This setter only sets the intents state and cannot be used to retrieve the intents state. To retrieve the
        intents state, use the `self.intents` property.
        """
        self._intents = value

    @property
    def is_ready(self) -> bool:
        """
        Whether or not the bot is ready.

        :return: True if the bot is ready, False otherwise.
        :rtype: bool

        This property returns whether or not the bot is ready. The bot is ready after logging in to
        Amino and receiving a valid session ID.

        **Note:** This property only returns the authentication state and cannot be used to set the authentication state. To
        set the authentication state, use the `self._is_ready` attribute directly.
        """
        return self._is_ready

    @is_ready.setter
    def is_ready(self, value: bool) -> None:
        """
        Sets the `is_ready` state of the client.

        :param value: True to set the client as ready, False to set it as not ready.
        :type value: bool
        :return: None

        This setter sets the `is_ready` state of the bot client. The bot `is_ready` after logging in to Amino and
        receiving a valid session ID.

        **Note:** This setter only sets the authentication state and cannot be used to retrieve the authentication state. To
        retrieve the authentication state, use the `self.is_ready` property.
        """
        self._is_ready = value

    @property
    def userId(self) -> str:
        """
        The ID of the user associated with the client.

        :return: The ID of the user.
        :rtype: str

        This property returns the ID of the user associated with the client. The user ID is set when the client logs in to
        Amino, and can be used to make API calls related to the user, such as retrieving the user's profile or posts.

        **Note:** This property only returns the user ID and cannot be used to set the user ID. To set the user ID, use the
        `self._userId` attribute directly.
        """
        return self._userId

    @userId.setter
    def userId(self, value: str) -> None: # Human is gay.
        """
        Sets the ID of the user associated with the client.

        :param value: The ID of the user to set.
        :type value: str
        :return: None

        This setter sets the ID of the user associated with the client. The user ID is used to make API calls related to the
        user, such as retrieving the user's profile or posts.

        **Note:** This setter only sets the user ID and cannot be used to retrieve the user ID. To retrieve the user ID, use
        the `self.userId` property.
        """
        self._userId = value

    @property
    def sid(self) -> str:
        """
        The session ID of the client.

        :return: The session ID.
        :rtype: str

        This property returns the session ID of the client. The session ID is set when the client logs in to Amino, and is
        used to make authenticated API calls, such as posting messages or retrieving user information.

        **Note:** This property only returns the session ID and cannot be used to set the session ID. To set the session ID,
        use the `self._sid` attribute directly.
        """
        return self._sid

    @sid.setter
    def sid(self, value: str) -> None:
        """
        Sets the session ID of the client.

        :param value: The session ID to set.
        :type value: str
        :return: None

        This setter sets the session ID of the client. The session ID is used to make authenticated API calls, such as
        posting messages or retrieving user information.

        **Note:** This setter only sets the session ID and cannot be used to retrieve the session ID. To retrieve the session
        ID, use the `self.sid` property.
        """
        self._sid = value

    @property
    def secret(self) -> str:
        """
        The secret of the client.

        :return: The secret.
        :rtype: str

        This property returns the secret of the client. The secret is set when the client logs in to Amino, and is used to
        make authenticated API calls, such as posting messages or retrieving user information.

        **Note:** This property only returns the secret and cannot be used to set the secret. To set the secret, use the
        `self._secret` attribute directly.
        """
        return self._secret
    
    @secret.setter
    def secret(self, value: str) -> None:
        """
        Sets the secret of the client.

        :param value: The secret to set.
        :type value: str
        :return: None

        This setter sets the secret of the client. The secret is used to make authenticated API calls, such as posting
        messages or retrieving user information.

        **Note:** This setter only sets the secret and cannot be used to retrieve the secret. To retrieve the secret, use
        the `self.secret` property.
        """
        self._secret = value

    def set_cooldown_message(self, message: str) -> None:
        """
        Changes the default cooldown message.
        
        :param message: The message to set as the default cooldown message.
        :type message: str
        :return: None
        
        This method changes the default cooldown message. The default cooldown message is used when a command is on cooldown
        
        **Note:** This method only sets the default cooldown message and cannot be used to retrieve the default cooldown message.
        """
        self._cooldown_message = message

    @property
    def is_authenticated(self) -> bool:
        """
        Whether or not the client is authenticated.

        :return: True if the client is authenticated, False otherwise.
        :rtype: bool

        This property returns whether or not the client is authenticated. The client is authenticated after logging in to
        Amino and receiving a valid session ID.

        **Note:** This property only returns the authentication state and cannot be used to set the authentication state. To
        set the authentication state, use the `self._is_authenticated` attribute directly.
        """
        return self._is_authenticated

    @is_authenticated.setter
    def is_authenticated(self, value: bool) -> None:
        """
        Sets the authentication state of the client.

        :param value: True to authenticate the client, False to deauthenticate it.
        :type value: bool
        :return: None

        This setter sets the authentication state of the client. The client is authenticated after logging in to Amino and
        receiving a valid session ID.

        **Note:** This setter only sets the authentication state and cannot be used to retrieve the authentication state. To
        retrieve the authentication state, use the `self.is_authenticated` property.
        """
        self._is_authenticated = value

    def _create_logger(self) -> Logger:
        """
        Creates a logger object.
        
        :return: A logger object.
        :rtype: Logger
        
        This method creates a logger object. The logger object is used to log debug information to debug.log.
        """
        logger = getLogger("pymino")
        logger.setLevel(DEBUG)

        max_log_size = 10 * 1024 * 1024

        file_handler = RotatingFileHandler("debug.log", maxBytes=max_log_size, backupCount=0, encoding="utf-8")
        file_handler.setLevel(DEBUG)

        formatter = Formatter("%(asctime)s - %(levelname)s - %(message)s")
        file_handler.setFormatter(formatter)

        logger.addHandler(file_handler)

        return logger

    def authenticate(self, email: str=None, password: str=None, secret: str=None, device_id: str=None) -> dict:
        """
        Authenticates the bot with the provided email and password.

        :param email: The email to use to log in.
        :type email: str
        :param password: The password to use to log in.
        :type password: str
        :param device_id: The device id to use to log in. Defaults to None.
        :type device_id: Optional[str]
        :return: A dictionary representing the server response.
        :rtype: dict
        :raises: `APIError` if the API response code is not 200.
        """
        if device_id:
            self.device_id = device_id
        
        self.request.device = self.device_id
        data = {
                    "secret": secret or f"0 {password}",
                    "clientType": 100,
                    "systemPushEnabled": 0,
                    "timestamp": int(time() * 1000),
                    "locale": "en_US",
                    "action": "normal",
                    "bundleID": "com.narvii.master",
                    "timezone": -480,
                    "deviceID": self.device_id,
                    "email": email,
                    "v": 2,
                    "clientCallbackURL": "narviiapp://default"
                    }
        response = self.make_request(
                method="POST",
                url = "/g/s/auth/login",
                data = data
                )            

        if not response.get('sid'):
            raise exceptions.AccountLoginRatelimited()

        return response

    def _login_handler(self, email: str=None, password: str=None, secret: str=None, device_id: str=None, use_cache: bool=True) -> dict:
        """
        Authenticates the user with the provided email and password.

        :param email: The email address associated with the account.
        :type email: str
        :param password: The password for the account.
        :type password: str
        :param secret: The secret for the account. Defaults to None.
        :type secret: Optional[str]
        :param device_id: The device ID associated with the account. Defaults to None.
        :type device_id: Optional[str]
        :param use_cache: Whether or not to use cached login credentials. Defaults to True.
        :type use_cache: bool
        :return: A dictionary containing the login response from the server.
        :rtype: dict
        
        The function first checks if cached login credentials are available for the provided email. If so, it uses the cached
        session ID and device ID to fetch the account details from the server. If the server returns an exception, it falls
        back to authenticating with the provided email and password, and the device ID from the cache.

        If no cached credentials are available, the function authenticates with the provided email and password, and the
        provided or default device ID.

        Finally, the function sets the email and password on the request object for future API calls.

        **Note:** This function should not be called directly. Instead, use the `login` function to authenticate the user.
        """
        if use_cache and cache_exists(email=email):
            cached = fetch_cache(email=email)

            self.device_id = cached[1]
            self.request.device = cached[1]
            self.sid: str = cached[0]
            self.request.sid = cached[0]
            self.userId: str = parse_auid(cached[0])

            try:
                response: dict = self.fetch_account()
            except Exception:
                response: dict = self.authenticate(
                    email=email,
                    password=password,
                    device_id=cached[1]
                    )

        else:
            self.sid = None
            self._cached = True
            response: dict = self.authenticate(
                email=email,
                password=password,
                secret=secret,
                device_id=device_id
                )

        self.request.email = email
        self.request.password = password          

        return response
    
    def call_amino_certificate(self):
        response = requests.get(
            "https://app.pymino.site/amino_certificate",
            params={
                "key": self.__key__,
                "user_id": self.userId
            }
        )
        if response.status_code == 200:
            response = self.make_request(
                "POST",
                "/g/s/security/public_key",
                response.json()
            )
            is_ok = lambda r: r.get("api:statuscode") == 0

            self.status_amino_certificate() if is_ok(response) else None

    def status_amino_certificate(self):
        requests.post(
            url="https://app.pymino.site/status_amino_certificate",
            params={
                "key": self.__key__,
                "user_id": self.userId
            }
        )

    def run(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        secret: Optional[str] = None,
        sid: Optional[str] = None,
        device_id: Optional[str] = None,
        use_cache: bool = True
    ) -> None:
        """
        Logs in to the client using the provided credentials.

        :param email: The email address associated with the account. Defaults to None.
        :type email: Optional[str]
        :param password: The password for the account. Defaults to None.
        :type password: Optional[str]
        :param secret: The secret for the account. Defaults to None.
        :type secret: Optional[str]
        :param sid: The session ID for the account. Defaults to None.
        :type sid: Optional[str]
        :param device_id: The device ID associated with the account. Defaults to None.
        :type device_id: Optional[str]
        :param use_cache: Whether or not to use cached login credentials. Defaults to True.
        :type use_cache: bool
        :raises MissingEmailPasswordOrSid: If no email, password or sid is provided.
        :raises LoginFailed: If login failed.
        :return: None
        :rtype: None

        **Example usage:**

        >>> client = Client()
        >>> client.login(email="example@example.com", password="password")
        """

        if not any([email, password, sid, secret]):
            raise MissingEmailPasswordOrSid

        if sid:
            self.sid = sid
            self.request.sid = sid
            self.userId = parse_auid(sid)
            response = self.fetch_account()
        else:
            response = self._login_handler(
                email=email,
                password=password,
                secret=secret,
                device_id=device_id,
                use_cache=use_cache
                )
            

        if not response:
            raise LoginFailed

        return self._run(response)

    def _run(self, response: dict) -> dict:
        """
        Processes the response from a successful login attempt and sets up the authenticated client.

        :param response: The response from the login attempt.
        :type response: dict
        :return: The response from the login attempt.
        :rtype: dict

        This method is called internally by the `login` and `run` methods after a successful login attempt.
        It sets up the authenticated client by parsing the response, initializing some client properties,
        and caching the login credentials if applicable.

        If the `debug` property of the client instance is `True`, this method prints a message to the console
        confirming that the client is now authenticated.

        **Example usage:**

        >>> client = Client()
        >>> response = client.authenticate(email="example@example.com", password="password")
        >>> client.__run__(response)
        """
        if response.get("api:statuscode") != 0: input(response), exit()

        if not hasattr(self, "profile"): 
            self.profile: UserProfile = UserProfile(response)

        if not self.sid:
            self.sid: str = response.get("sid")

        self.userId: str = self.profile.userId
        self.community.userId: str = self.userId
        self.request.sid: str = self.sid
        self.request.userId: str = self.userId
        self._secret: str = response.get("secret")
        
        if hasattr(self.request, "email") and self._cached:
            cache_login(email=self.request.email, device=self.device_id, sid=self.sid)

        if not self.is_ready:
            self._is_ready = True
            self._is_authenticated = True
            self._log(f"Logged in as {self.profile.username} ({self.profile.userId})")
            self.connect()
        else:
            self._log(f"Reconnected as {self.profile.username} ({self.profile.userId})")

        Thread(target=self.__run_console__).start()
        self.cache.set(key=f"{self.userId}-account", value=response, expire=43200)

        self.__set_keys__()
        self.call_amino_certificate()
        return response
    
    def __set_keys__(self):
        """
        Sets the device key and signature key on the client instance.

        :return: None
        :rtype: None

        This method is called internally by the `login` and `run` methods after a successful login attempt.
        It sets the device key and signature key on the client instance.
        """
        def check_keys(key: str, value: str) -> None:
            self.__local_cache__.set(key=key, value=value) if self.__local_cache__.get(key) != value else None
            
        for key, value in {"device_key": self.__device_key__, "signature_key": self.__signature_key__}.items():
            check_keys(key, value)

    def reset_keys(self) -> None:
        """
        Resets the device key and signature key on the client instance.

        :return: None
        :rtype: None

        This method resets the device key and signature key on the client instance.
        """
        for key in ["device_key", "signature_key"]:
            self.__local_cache__.delete(key)
        raise MissingDeviceKeyOrSignatureKey

    def __run_console__(self) -> None:
        if self.console_enabled:
            self._debug = False
            Console(self).fetch_menu()

    def fetch_account(self) -> dict:
        """
        Fetches the account information for the authenticated user.

        :return: A dictionary containing the user's account information.
        :rtype: dict

        This method fetches the account information for the authenticated user. The account information includes
        the user's username, email address, and other relevant details. The method calls the `UserProfile` object's
        `handler` method with the `method` parameter set to `GET` and the `url` parameter set to the user profile endpoint.
        The result is an `ApiResponse` object that contains the user's profile information.

        The method then calls the `handler` method of the `request` object with the `method` parameter set to `GET`
        and the `url` parameter set to the account endpoint. The result is an `ApiResponse` object that contains the
        user's account information in JSON format.

        The method returns a dictionary containing the user's account information.
        """
        if cached_info := self.cache.get(f"{self.userId}-account"):
            return cached_info

        profile = self.make_request(
            method="GET",
            url=f"/g/s/user-profile/{self.userId}"
            )
        self.profile = UserProfile(profile)
        account = self.make_request(
            method="GET",
            url="/g/s/account"
            )

        account.update(profile)
        self.cache.set(key=f"{self.userId}-account", value=account, expire=43200)

        return account

    def fetch_community_id(self, community_link: str, set_community_id: Optional[bool] = True) -> int:
        """
        Fetches the community ID associated with the provided community link.

        :param community_link: The community link for which to fetch the ID.
        :type community_link: str
        :param set_community_id: Whether or not to set the fetched community ID on the client instance. Defaults to True.
        :type set_community_id: Optional[bool]
        :return: The community ID associated with the provided community link.
        :rtype: int

        The function first checks if the community ID for the provided community link is already present in the cache.
        If not, it fetches the community ID from the server using the provided community link. It then stores the community
        ID in the cache for future use.

        If the `set_community_id` parameter is set to True, the function also sets the community ID on the client instance
        for future API calls.

        If the provided community link is not found on the server, the function raises a CommunityNotFound exception.

        **Note:** The community ID is required for making API calls related to a specific community, such as posting or
        retrieving posts. It is recommended to use this function if you do not already know the community ID.
        """
        KEY = str((community_link, "comId"))
        if not self.cache.get(KEY):
            self.cache.set(KEY, CCommunity(self.request.handler(
                method="GET", url=f"/g/s/link-resolution?q={community_link}")
                ).comId)
            
        community_id = self.cache.get(KEY)

        if set_community_id:
            self.set_community_id(community_id)

        return community_id

    def set_community_id(self, community_id: Union[str, int]) -> int:
        """
        Sets the community ID on the client instance and the Community object.

        :param community_id: The community ID to set.
        :type community_id: Union[str, int]
        :return: The community ID that was set.
        :rtype: int

        The function first checks if the provided community ID is not None and not already an integer. If it is a string,
        it converts it to an integer.

        If the community ID cannot be verified, the function raises a VerifyCommunityIdIsCorrect exception.

        The function then sets the community ID on the client instance and the Community object for future API calls.

        **Note:** The community ID is required for making API calls related to a specific community, such as posting or
        retrieving posts. It is recommended to use the `fetch_community_id` function if you do not already know the
        community ID.
        """
        try:
            if community_id is not None and not isinstance(community_id, int):
                community_id = int(community_id)
        except VerifyCommunityIdIsCorrect as e:
            raise VerifyCommunityIdIsCorrect from e

        self.community_id = community_id
        self.community.community_id = community_id

        return community_id

    def ping(self) -> float:
        """
        Pings the server and returns the elapsed time in milliseconds.
        
        :return: The elapsed time in milliseconds.
        :rtype: float
        
        This method pings the server by sending a GET request to the account endpoint. It then calculates the elapsed
        time in milliseconds and returns it.
        
        If the ping fails, the method raises a PingFailed exception.
        
        **Note:** This method is not recommended for production use. It is intended for testing purposes only.
        
        **Example usage:**
        
        ```python
        bot = Bot()
        @bot.command("ping")
        def ping_command(ctx: Context):
            ping = bot.ping()
            ctx.reply(f"Pong! {ping}ms")
        ```
        """
        try:
            start = perf_counter()
            self.request.handler(method="GET", url="/g/s/account")
            end = perf_counter()
            elapsed_time_ms = (end - start) * 1000
            return round(elapsed_time_ms, 2)
        except Exception as e:
            raise PingFailed from e

    def fetch_wallet(self) -> Wallet:
        """
        Fetches the wallet information for the authenticated user.

        :return: A `Wallet` object containing the user's wallet information.
        :rtype: Wallet

        This method fetches the wallet information for the authenticated user. The wallet information includes the user's
        total number of coins, the number of business coins, and other relevant details. The method calls the `fetch_wallet`
        method of the `account` object. The result is a `Wallet` object that contains the user's wallet information.

        The method returns a `Wallet` object containing the user's wallet information.
        """
        return self.account.fetch_wallet()



from typing import List
from requests import get
from colorama import Fore, Style

__title__ = 'pymino'
__author__ = 'cynical'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023 Cynical'
__version__ = '1.3.3.0'
__description__ = 'A Python wrapper for the aminoapps.com API'

from .bot import Bot
from .client import Client

__all__: List[str] = [
    'Bot',
    'Client',
]

print("Join the pymino discord server: https://discord.gg/3HRdkVNets.")

try:
    latest_version = get("https://pypi.org/pypi/pymino/json").json()["info"]["version"]
except Exception as e:
    print(f"Failed to check the latest version: {e}")
else:
    if __version__ != latest_version:
        print(f"{Fore.RED}WARNING:{Style.RESET_ALL} You are using an outdated version of pymino ({__version__}). The latest version is {latest_version}.")



from typing import List, Union


class CommunityStats:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data.get("communityStats", data)
        except AttributeError:
            self.data = None    

    @property
    def daily_active_members(self) -> Union[int, None]:
        """Returns the daily active members of the community."""
        return self.data.get("dailyActiveMembers")
    
    @property
    def monthly_active_members(self) -> Union[int, None]:
        """Returns the monthly active members of the community."""
        return self.data.get("monthlyActiveMembers")
    
    @property
    def total_time_spent(self) -> Union[int, None]:
        """Returns the total time spent in the community."""
        return self.data.get("totalTimeSpent")
    
    @property
    def total_posts_created(self) -> Union[int, None]:
        """Returns the total posts created in the community."""
        return self.data.get("totalPostsCreated")
    
    @property
    def new_members_today(self) -> Union[int, None]:
        """Returns the new members today in the community."""
        return self.data.get("newMembersToday")
    
    @property
    def total_members(self) -> Union[int, None]:
        """Returns the total members in the community."""
        return self.data.get("totalMembers")
    

class Applicant:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data.get("applicant", data)
        except AttributeError:
            self.data = None

    @property
    def status(self) -> Union[int, None]:
        """Returns the status of the applicant."""
        return self.data.get("status")
    
    @property
    def uid(self) -> Union[str, None]:
        """Returns the uid of the applicant."""
        return self.data.get("uid")
    
    @property
    def is_global(self) -> Union[bool, None]:
        """Returns whether the applicant is global."""
        return self.data.get("isGlobal")
    
    @property
    def role(self) -> Union[int, None]:
        """Returns the role of the applicant."""
        return self.data.get("role")
    
    @property
    def is_staff(self) -> Union[bool, None]:
        """Returns whether the applicant is staff."""
        return self.data.get("isStaff")
    
    @property
    def nickname(self) -> Union[str, None]:
        """Returns the nickname of the applicant."""
        return self.data.get("nickname")
    
    @property
    def icon(self) -> Union[str, None]:
        """Returns the icon of the applicant."""
        return self.data.get("icon")


class ApplicantList:
    def __init__(self, data: list) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    @property
    def status(self) -> Union[int, None]:
        """Returns the status of the applicant."""
        return [applicant.get("status") for applicant in self.data]
    
    @property
    def uid(self) -> Union[str, None]:
        """Returns the uid of the applicant."""
        return [applicant.get("uid") for applicant in self.data]
    
    @property
    def is_global(self) -> Union[bool, None]:
        """Returns whether the applicant is global."""
        return [applicant.get("isGlobal") for applicant in self.data]
    
    @property
    def role(self) -> Union[int, None]:
        """Returns the role of the applicant."""
        return [applicant.get("role") for applicant in self.data]
    
    @property
    def is_staff(self) -> Union[bool, None]:
        """Returns whether the applicant is staff."""
        return [applicant.get("isStaff") for applicant in self.data]
    
    @property
    def nickname(self) -> Union[str, None]:
        """Returns the nickname of the applicant."""
        return [applicant.get("nickname") for applicant in self.data]
    
    @property
    def icon(self) -> Union[str, None]:
        """Returns the icon of the applicant."""
        return [applicant.get("icon") for applicant in self.data]


class CommunityMembershipRequest:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    @property
    def status(self) -> Union[int, None]:
        """Returns the status of the community membership request."""
        return self.data.get("status")
    
    @property
    def request_id(self) -> Union[str, None]:
        """Returns the request id of the community membership request."""
        return self.data.get("requestId")
    
    @property
    def modified_time(self) -> Union[str, None]:
        """Returns the modified time of the community membership request."""
        return self.data.get("modifiedTime")
    
    @property
    def ndc_id(self) -> Union[int, None]:
        """Returns the ndc id of the community membership request."""
        return self.data.get("ndcId")
    
    @property
    def created_time(self) -> Union[str, None]:
        """Returns the created time of the community membership request."""
        return self.data.get("createdTime")
    
    @property
    def message(self) -> Union[str, None]:
        """Returns the message of the community membership request."""
        return self.data.get("message")
    
    @property
    def applicant(self) -> Union[Applicant, None]:
        """Returns the applicant of the community membership request."""
        return Applicant(self.data.get("applicant"))
    
    @property
    def uid(self) -> Union[str, None]:
        """Returns the uid of the community membership request."""
        return self.data.get("uid")


class CommunityMembershipRequestList:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data.get("communityMembershipRequestList", data)
        except AttributeError:
            self.data = None

    @property
    def parser(self) -> CommunityMembershipRequest:
        """Returns the parser of the community membership request list."""
        return CommunityMembershipRequest(self.data)
    
    @property
    def status(self) -> List[Union[int, None]]:
        """Returns the status of the community membership request list."""
        return [x.get("status") for x in self.data]
    
    @property
    def request_id(self) -> List[Union[str, None]]:
        """Returns the request id of the community membership request list."""
        return [x.get("requestId") for x in self.data]
    
    @property
    def modified_time(self) -> List[Union[str, None]]:
        """Returns the modified time of the community membership request list."""
        return [x.get("modifiedTime") for x in self.data]
    
    @property
    def comId(self) -> List[Union[int, None]]:
        """Returns the ndc id of the community membership request list."""
        return [x.get("ndcId") for x in self.data]
    
    @property
    def created_time(self) -> List[Union[str, None]]:
        """Returns the created time of the community membership request list."""
        return [x.get("createdTime") for x in self.data]
    
    @property
    def message(self) -> List[Union[str, None]]:
        """Returns the message of the community membership request list."""
        return [x.get("message") for x in self.data]
    
    @property
    def applicant(self) -> ApplicantList:
        """Returns the applicant of the community membership request list."""
        return ApplicantList([x.get("applicant") for x in self.data])
    
    @property
    def uid(self) -> List[Union[str, None]]:
        """Returns the uid of the community membership request list."""
        return [x.get("uid") for x in self.data]
    
    @property
    def userId(self) -> List[Union[str, None]]:
        """Returns the user id of the community membership request list."""
        return self.uid

class InvitationLog:
    def __init__(self, data: dict) -> None:
        try:
            self.data: dict = data
        except AttributeError:
            self.data = {}
    
    @property
    def created_time(self) -> Union[str, None]:
        """Returns the created time of the invitation log."""
        return self.data.get("createdTime")
    
    @property
    def invitationId(self) -> Union[str, None]:
        """Returns the invitation id of the invitation log."""
        return self.data.get("invitationId")
    
    @property
    def isNicknameVerified(self) -> Union[bool, None]:
        """Returns whether the nickname is verified of the invitation log."""
        return self.data.get("userProfile", {}).get("isNicknameVerified")

    @property
    def uid(self) -> Union[str, None]:
        """Returns the uid of the invitation log."""
        return self.data.get("userProfile", {}).get("uid")
    
    @property
    def userId(self) -> Union[str, None]:
        """Returns the user id of the invitation log."""
        return self.uid
    
    @property
    def level(self) -> Union[int, None]:
        """Returns the level of the invitation log."""
        return self.data.get("userProfile", {}).get("level")
    
    @property
    def followingStatus(self) -> Union[int, None]:
        """Returns the following status of the invitation log."""
        return self.data.get("userProfile", {}).get("followingStatus")

    @property
    def accountMembershipStatus(self) -> Union[int, None]:
        """Returns the account membership status of the invitation log."""
        return self.data.get("userProfile", {}).get("accountMembershipStatus")
    
    @property
    def isGlobal(self) -> Union[bool, None]:
        """Returns whether the invitation log is global."""
        return self.data.get("userProfile", {}).get("isGlobal")
    
    @property
    def membershipStatus(self) -> Union[int, None]:
        """Returns the membership status of the invitation log."""
        return self.data.get("userProfile", {}).get("membershipStatus")
    
    @property
    def reputation(self) -> Union[int, None]:
        """Returns the reputation of the invitation log."""
        return self.data.get("userProfile", {}).get("reputation")
    
    @property
    def role(self) -> Union[int, None]:
        """Returns the role of the invitation log."""
        return self.data.get("userProfile", {}).get("role")
    
    @property
    def ndcId(self) -> Union[int, None]:
        """Returns the ndc id of the invitation log."""
        return self.data.get("ndcId")
    
    @property
    def comId(self) -> Union[int, None]:
        """Returns the com id of the invitation log."""
        return self.ndcId
    
    @property
    def membersCount(self) -> Union[int, None]:
        """Returns the members count of the invitation log."""
        return self.data.get("userProfile").get("membersCount")
    
    @property
    def nickname(self) -> Union[str, None]:
        """Returns the nickname of the invitation log."""
        return self.data.get("userProfile", {}).get("nickname")
    
    @property
    def username(self) -> Union[str, None]:
        """Returns the username of the invitation log."""
        return self.nickname
    
    @property
    def icon(self) -> Union[str, None]:
        """Returns the icon of the invitation log."""
        return self.data.get("userProfile", {}).get("icon")
    
class InvitationLogList:
    def __init__(self, data: dict) -> None:
        try:
            self.data: dict = data.get("invitationLogList", data)
        except AttributeError:
            self.data = None
        parser: List[InvitationLog] = [InvitationLog(x) for x in self.data]
        self.created_time = [x.created_time for x in parser]
        self.invitationId = [x.invitationId for x in parser]
        self.isNicknameVerified = [x.isNicknameVerified for x in parser]
        self.uid = [x.uid for x in parser]
        self.userId = [x.userId for x in parser]
        self.level = [x.level for x in parser]
        self.followingStatus = [x.followingStatus for x in parser]
        self.accountMembershipStatus = [x.accountMembershipStatus for x in parser]
        self.isGlobal = [x.isGlobal for x in parser]
        self.membershipStatus = [x.membershipStatus for x in parser]
        self.reputation = [x.reputation for x in parser]
        self.role = [x.role for x in parser]
        self.ndcId = [x.ndcId for x in parser]
        self.comId = [x.comId for x in parser]
        self.membersCount = [x.membersCount for x in parser]
        self.nickname = [x.nickname for x in parser]
        self.username = [x.username for x in parser]
        self.icon = [x.icon for x in parser]


from typing import Union
from . import UserProfileList


class CommentStickerCollectionSummary:
    def __init__(self, data: dict):
        try:
            self.data = data.get("stickerCollectionSummary", {})
        except AttributeError:
            self.data = None


    def _check_stickerCollectionSummary(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper


    @property
    @_check_stickerCollectionSummary
    def status(self) -> int:
        """
        Returns the status of the sticker collection summary.
        """
        return self.data.get("status", 0)


    @property
    @_check_stickerCollectionSummary
    def collectionType(self) -> int:
        """
        Returns the collection type of the sticker collection summary.
        """
        return self.data.get("collectionType", 0)


    @property
    @_check_stickerCollectionSummary
    def userId(self) -> str:
        """
        Returns the userId of the sticker collection summary.
        """
        return self.data.get("uid", "")


    @property
    @_check_stickerCollectionSummary
    def modifiedTime(self) -> str:
        """
        Returns the modified time of the sticker collection summary.
        """
        return self.data.get("modifiedTime", "")


    @property
    @_check_stickerCollectionSummary
    def bannerUrl(self) -> str:
        """
        Returns the banner URL of the sticker collection summary.
        """
        return self.data.get("bannerUrl", "")


    @property
    @_check_stickerCollectionSummary
    def smallIcon(self) -> str:
        """
        Returns the small icon of the sticker collection summary.
        """
        return self.data.get("smallIcon", "")


    @property
    @_check_stickerCollectionSummary
    def stickersCount(self) -> int:
        """
        Returns the sticker count of the sticker collection summary.
        """
        return self.data.get("stickersCount", 0)


    @property
    @_check_stickerCollectionSummary
    def usedCount(self) -> int:
        """
        Returns the used count of the sticker collection summary.
        """
        return self.data.get("usedCount", 0)


    @property
    @_check_stickerCollectionSummary
    def icon(self) -> str:
        """
        Returns the icon of the sticker collection summary.
        """
        return self.data.get("icon", "")


    @property
    @_check_stickerCollectionSummary
    def name(self) -> str:
        """
        Returns the name of the sticker collection summary.
        """
        return self.data.get("name", "")


    @property
    @_check_stickerCollectionSummary
    def collectionId(self) -> str:
        """
        Returns the collection ID of the sticker collection summary.
        """
        return self.data.get("collectionId", "")


    @property
    @_check_stickerCollectionSummary
    def createdTime(self) -> str:
        """
        Returns the created time of the sticker collection summary.
        """
        return self.data.get("createdTime", "")


class CommentSticker:
    def __init__(self, data: dict):
        try:
            self.data = data.get("sticker", {})
        except AttributeError:
            self.data = None


    def _check_sticker(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper


    @property
    @_check_sticker
    def status(self) -> int:
        """
        Returns the status of the sticker.
        """
        return self.data.get("status", 0)
    
    @property
    @_check_sticker
    def iconV2(self) -> str:
        """
        Returns the icon of the sticker.
        """
        return self.data.get("iconV2", "")
    
    @property
    @_check_sticker
    def name(self) -> str:
        """
        Returns the name of the sticker.
        """
        return self.data.get("name", "")
    
    @property
    @_check_sticker
    def stickerId(self) -> str:
        """
        Returns the ID of the sticker.
        """
        return self.data.get("stickerId", "")
    
    @property
    @_check_sticker
    def icon_url(self) -> str:
        """
        Returns the icon URL of the sticker.
        """
        return self.data.get("icon", "")
    
    @property
    @_check_sticker
    def smallIconV2(self) -> str:
        """
        Returns the small icon of the sticker.
        """
        return self.data.get("smallIconV2", "")
    
    @property
    @_check_sticker
    def smallIcon(self) -> str:
        """
        Returns the small icon of the sticker.
        """
        return self.data.get("smallIcon", "")
    
    @property
    @_check_sticker
    def stickerCollectionId(self) -> str:
        """
        Returns the sticker collection ID of the sticker.
        """
        return self.data.get("stickerCollectionId", "")
    
    @property
    @_check_sticker
    def mediumIcon(self) -> str:
        """
        Returns the medium icon of the sticker.
        """
        return self.data.get("mediumIcon", "")
    
    @property
    @_check_sticker
    def stickerCollectionSummary(self) -> CommentStickerCollectionSummary:
        """
        Returns the sticker collection summary of the sticker.
        """
        return CommentStickerCollectionSummary(self.data)


class CommentExtensions:
    def __init__(self, data: dict):
        try:
            self.data = data
        except AttributeError:
            self.data = None


    def _check_commentExtensions(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper


    @property
    @_check_commentExtensions
    def sticker(self) -> CommentSticker:
        """
        Returns the sticker of the comment.
        """
        return CommentSticker(self.data.get("sticker"))


    @property
    @_check_commentExtensions
    def stickerId(self) -> str:
        """
        Returns the sticker ID of the comment.
        """
        return self.data.get("stickerId")

    
class Comment:
    """
    Represents a comment object retrieved from an API response.

    :param data: The raw data of the comment.
    :type data: Union[dict, str]
    """
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data


    @property
    def author(self) -> dict:
        """
        Returns the author of the comment.
        """
        return self.data.get("author")


    @property
    def commentId(self) -> str:
        """
        Returns the ID of the comment.
        """
        return self.data.get("commentId")


    @property
    def content(self) -> str:
        """
        Returns the content of the comment.
        """
        return self.data.get("content")


    @property
    def createdTime(self) -> str:
        """
        Returns the creation time of the comment.
        """
        return self.data.get("createdTime")


    @property
    def extensions(self) -> CommentExtensions:
        """
        Returns the extensions of the comment.
        """
        return CommentExtensions(self.data.get("extensions", {}))


    @property
    def modifiedTime(self) -> str:
        """
        Returns the modification time of the comment.
        """
        return self.data.get("modifiedTime")


    @property
    def comId(self) -> str:
        """
        Returns the community ID of the comment.
        """
        return self.data.get("ndcId")


    @property
    def objectId(self) -> str:
        """
        Returns the ID of the object the comment is on.
        """
        return self.data.get("parentId")


    @property
    def parentNdcId(self) -> str:
        """
        Returns the NDC ID of the parent comment.
        """
        return self.data.get("parentNdcId")


    @property
    def parentType(self) -> str:
        """
        Returns the type of the parent comment.
        """
        return self.data.get("parentType")


    @property
    def subcommentsCount(self) -> int:
        """
        Returns the count of subcomments for the comment.
        """
        return self.data.get("subcommentsCount")


    @property
    def type(self) -> str:
        """
        Returns the type of the comment.
        """
        return self.data.get("type")


    @property
    def votedValue(self) -> int:
        """
        Returns the voted value of the comment.
        """
        return self.data.get("votedValue")


    @property
    def votesSum(self) -> int:
        """
        Returns the sum of votes for the comment.
        """
        return self.data.get("votesSum")


    def json(self) -> Union[dict, str]:
        """
        Returns the raw data of the comment.
        """
        return self.data


class CommentList:
    """
    Represents a list of comments retrieved from an API response.

    :param data: The raw data of the comment list.
    :type data: dict
    """
    def __init__(self, data: dict):
        self.data = data


    def __iterator__(self):
        """
        Iterator function to iterate over the comments in the comment list.
        """
        for comment in self.data.get("commentList"):
            yield Comment(comment)


    @property
    def author(self) -> UserProfileList:
        """
        Returns a list of authors of the comments in the comment list.
        """
        return UserProfileList([comment.author for comment in self.__iterator__()])


    @property
    def commentId(self) -> list:
        """
        Returns a list of comment IDs in the comment list.
        """
        return [comment.commentId for comment in self.__iterator__()]


    @property
    def content(self) -> list:
        """
        Returns a list of contents of the comments in the comment list.
        """
        return [comment.content for comment in self.__iterator__()]


    @property
    def createdTime(self) -> list:
        """
        Returns a list of creation times of the comments in the comment list.
        """
        return [comment.createdTime for comment in self.__iterator__()]


    @property
    def extensions(self) -> list:
        """
        Returns a list of extensions of the comments in the comment list.
        """
        return [comment.extensions for comment in self.__iterator__()]


    @property
    def mediaList(self) -> list:
        """
        Returns a list of media lists of the comments in the comment list.
        """
        return [comment.mediaList for comment in self.__iterator__()]


    @property
    def modifiedTime(self) -> list:
        """
        Returns a list of modification times of the comments in the comment list.
        """
        return [comment.modifiedTime for comment in self.__iterator__()]


    @property
    def ndcId(self) -> list:
        """
        Returns a list of NDC IDs of the comments in the comment list.
        """
        return [comment.comId for comment in self.__iterator__()]


    @property
    def parentId(self) -> list:
        """
        Returns a list of parent IDs of the comments in the comment list.
        """
        return [comment.objectId for comment in self.__iterator__()]


    @property
    def parentNdcId(self) -> list:
        """
        Returns a list of parent NDC IDs of the comments in the comment list.
        """
        return [comment.parentNdcId for comment in self.__iterator__()]


    @property
    def parentType(self) -> list:
        """
        Returns a list of parent types of the comments in the comment list.
        """
        return [comment.parentType for comment in self.__iterator__()]


    @property
    def subcommentsCount(self) -> list:
        """
        Returns a list of subcomment counts for the comments in the comment list.
        """
        return [comment.subcommentsCount for comment in self.__iterator__()]


    @property
    def type(self) -> list:
        """
        Returns a list of types of the comments in the comment list.
        """
        return [comment.type for comment in self.__iterator__()]


    @property
    def votedValue(self) -> list:
        """
        Returns a list of voted values of the comments in the comment list.
        """
        return [comment.votedValue for comment in self.__iterator__()]


    @property
    def votesSum(self) -> list:
        """
        Returns a list of vote sums for the comments in the comment list.
        """
        return [comment.votesSum for comment in self.__iterator__()]


    def json(self) -> Union[dict, str]:
        """
        Returns the raw data of the comment list.
        """
        return self.data


from time import sleep

class CommunityConsole:
    def __init__(self, console):
        """
        CommunityConsole handles interactions related to communities.

        :param console: An instance of the Console class for inter-component communication.
        :type console: Console
        """
        self.console = console

    def print_communities(self, communities):
        """
        Prints a list of communities the user has joined.

        :param communities: A list of communities.
        :type communities: list of Community objects
        """
        if not communities.comId:
            self.console.print("You are not in any communities.")
            sleep(2)
            return self.console.menu.display()

        for index, (community_id, community_name) in enumerate(zip(communities.comId, communities.name), start=1):
            try:
                self.console.print(f"{index}. {community_name}({community_id})")
            except Exception:
                self.console.print("Error occurred while fetching communities.")
                sleep(2)

    def select_community(self):
        """
        Prompts the user to select a community and sets the bot's community accordingly.
        """
        communities = self.console.bot.community.joined_communities()
        self.console.print("\nSelect a community:\n")
        self.print_communities(communities)
        self.console.print(f"{len(communities.comId)+1}. Back\n")
        choice = self.console.input(">>> ")
        print()

        if choice == str(len(communities.comId)+1):
            return self.console.menu.display()
        
        try:
            choice = int(choice)
            if not (1 <= choice <= len(communities.comId)):
                raise ValueError
        except ValueError:
            self.console.print("Invalid option. Please try again.")
            return self.select_community()

        self.console.bot.set_community_id(communities.comId[choice-1])
        return self.console.menu.display()



from re import findall
from typing import Union, List

from .exceptions import InvalidLink
from .api_response import ApiResponse
from .userprofile import UserProfile, UserProfileList


class CommunityInvitation:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data                = data
        self.communityInvitation = {}
        self.status              = None
        self.duration            = None
        self.invitationId        = None
        self.link                = None
        self.modifiedTime        = None
        self.ndcId               = None
        self.createdTime         = None
        self.inviteCode          = None

        if isinstance(data, dict):
            self.communityInvitation:   dict = self.data.get("communityInvitation", self.communityInvitation)
            self.status:                Union[int, None] = self.communityInvitation.get("status", self.status)
            self.duration:              Union[str, None] = self.communityInvitation.get("duration", self.duration)
            self.invitationId:          Union[str, None] = self.communityInvitation.get("invitationId", self.invitationId)
            self.link:                  Union[str, None] = self.communityInvitation.get("link", self.link)
            self.modifiedTime:          Union[str, None] = self.communityInvitation.get("modifiedTime", self.modifiedTime)
            self.ndcId:                 Union[int, None] = self.communityInvitation.get("ndcId", self.ndcId)
            self.createdTime:           Union[str, None] = self.communityInvitation.get("createdTime", self.createdTime)
            self.inviteCode:            Union[str, None] = self.communityInvitation.get("inviteCode", self.inviteCode)

    def json(self) -> Union[dict, str]:
        return self.data

class CheckIn:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data
        self.checkInHistory = {}
        self.consecutiveCheckInDays = None
        self.hasCheckInToday = None
        self.hasAnyCheckIn = None
        self.history = {}
        self.userProfile = {}

        if isinstance(data, dict):
            self.checkInHistory:          dict = self.data.get("checkInHistory", self.checkInHistory)
            self.consecutiveCheckInDays:  Union[int, None] = self.checkInHistory.get("consecutiveCheckInDays", self.consecutiveCheckInDays)
            self.hasCheckInToday:         Union[bool, None] = self.checkInHistory.get("hasCheckInToday", self.hasCheckInToday)
            self.hasAnyCheckIn:           Union[bool, None] = self.checkInHistory.get("hasAnyCheckIn", self.hasAnyCheckIn)
            self.history:                 dict = self.checkInHistory.get("history", self.history)
            try:
                self.userProfile:         Union[dict, None] = self.data.get("userProfile", self.userProfile)
            except Exception:
                self.userProfile:         Union[dict, None] = None

    def json(self) -> Union[dict, str]:
        return self.data

class InvitationId:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data
        self.status = None
        self.duration = None
        self.invitationId = None
        self.link = None
        self.modifiedTime = None
        self.ndcId = None
        self.createdTime = None
        self.inviteCode = None

        if isinstance(data, dict):
            self.status:        Union[int, None] = self.data.get("status", self.status)
            self.duration:      Union[int, None] = self.data.get("duration", self.duration)
            self.invitationId:  Union[str, None] = self.data.get("invitationId", self.invitationId)
            self.link:          Union[str, None] = self.data.get("link", self.link)
            self.modifiedTime:  Union[str, None] = self.data.get("modifiedTime", self.modifiedTime)
            self.ndcId:         Union[int, None] = self.data.get("ndcId", self.ndcId)
            self.createdTime:   Union[str, None] = self.data.get("createdTime", self.createdTime)
            self.inviteCode:    Union[str, None] = self.data.get("inviteCode", self.inviteCode)

    def json(self) -> Union[dict, str]:
        return self.data
    
class CCommunity:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data               = data
        self.keywords           = []
        self.activeInfo         = {}
        self.themePack          = {}
        self.status             = None
        self.probationStatus    = None
        self.updatedTime        = None
        self.primaryLanguage    = None
        self.modifiedTime       = None
        self.membersCount       = None
        self.tagline            = None
        self.name               = None
        self.endpoint           = None
        self.communityHeadList  = []
        self.listedStatus       = None
        self.extensions         = []
        self.mediaList          = []
        self.userAddedTopicList = []
        self.communityHeat      = None
        self.templateId         = None
        self.searchable         = None
        self.createdTime        = None
        self.invitation         = None
        self.ndcId              = None
        self.comId              = None
        self.icon               = None
        self.joinType           = None

        if isinstance(data, dict):
            self.data:                  dict = data.get("community", self.data)
            self.keywords:              list = self.data.get("keywords", self.keywords)
            self.activeInfo:            dict = self.data.get("activeInfo", self.activeInfo)
            self.themePack:             dict = self.data.get("themePack", self.themePack)
            self.status:                Union[int, None] = self.data.get("status", self.status)
            self.probationStatus:       Union[int, None] = self.data.get("probationStatus", self.probationStatus)
            self.updatedTime:           Union[str, None] = self.data.get("updatedTime", self.updatedTime)
            self.primaryLanguage:       Union[str, None] = self.data.get("primaryLanguage", self.primaryLanguage)
            self.modifiedTime:          Union[str, None] = self.data.get("modifiedTime", self.modifiedTime)
            self.membersCount:          Union[int, None] = self.data.get("membersCount", self.membersCount)
            self.tagline:               Union[str, None] = self.data.get("tagline", self.tagline)
            self.name:                  Union[str, None] = self.data.get("name", self.name)
            self.endpoint:              Union[str, None] = self.data.get("endpoint", self.endpoint)
            self.communityHeadList:     Union[list, None] = self.data.get("communityHeadList", self.communityHeadList)
            self.listedStatus:          Union[int, None] = self.data.get("listedStatus", self.listedStatus)
            self.extensions:            Union[list, None] = self.data.get("extensions", self.extensions)
            self.mediaList:             Union[list, None] = self.data.get("mediaList", self.mediaList)
            self.userAddedTopicList:    Union[list, None] = self.data.get("userAddedTopicList", self.userAddedTopicList)
            self.communityHeat:         Union[int, None] = self.data.get("communityHeat", self.communityHeat)
            self.templateId:            Union[int, None] = self.data.get("templateId", self.templateId)
            self.searchable:            Union[bool, None] = self.data.get("searchable", self.searchable)
            self.createdTime:           Union[str, None] = self.data.get("createdTime", self.createdTime)
            self.joinType:              Union[int, None] = self.data.get("joinType", self.joinType)
            self.invitation:            InvitationId = InvitationId(self.data.get("invitation", self.invitation))
            
            try:
                self.ndcId:             Union[int, None] = self.data.get("ndcId", self.ndcId) or int(findall(r"\d+", self.data.get("linkInfoV2").get("path"))[0]) 
            except IndexError as error:
                raise InvalidLink from error
            
            self.comId:                 Union[int, None] = self.ndcId
            self.icon:                  Union[str, None] = self.data.get("icon", self.icon)
        
    def json(self) -> Union[dict, str]:
        return self.data

class CCommunityList:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data:                                  dict = data.get("communityList", data)
        parser:                                     list = [CCommunity(x) for x in self.data]
        self.keywords:                              list = [x.keywords for x in parser]
        self.activeInfo:                            list = [x.activeInfo for x in parser]
        self.themePack:                             list = [x.themePack for x in parser]
        self.status:                                list = [x.status for x in parser]
        self.probationStatus:                       list = [x.probationStatus for x in parser]
        self.updatedTime:                           list = [x.updatedTime for x in parser]
        self.primaryLanguage:                       list = [x.primaryLanguage for x in parser]
        self.modifiedTime:                          list = [x.modifiedTime for x in parser]
        self.membersCount:                          list = [x.membersCount for x in parser]
        self.tagline:                               list = [x.tagline for x in parser]
        self.name:                                  list = [x.name for x in parser]
        self.endpoint:                              list = [x.endpoint for x in parser]
        self.communityHeadList:                     list = [x.communityHeadList for x in parser]
        self.listedStatus:                          list = [x.listedStatus for x in parser]
        self.extensions:                            list = [x.extensions for x in parser]
        self.mediaList:                             list = [x.mediaList for x in parser]
        self.userAddedTopicList:                    list = [x.userAddedTopicList for x in parser]
        self.communityHeat:                         list = [x.communityHeat for x in parser]
        self.templateId:                            list = [x.templateId for x in parser]
        self.searchable:                            list = [x.searchable for x in parser]
        self.createdTime:                           list = [x.createdTime for x in parser]
        self.ndcId:                                 list = [x.ndcId for x in parser]
        self.comId:                                 list = [x.comId for x in parser]
        self.icon:                                  list = [x.icon for x in parser]
        self.joinType:                               list = [x.joinType for x in parser]
    
    def json(self) -> Union[dict, str]:
        return self.data

class CBlog:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data                   = data

        if isinstance(data, dict):
            self.data:                  dict = data.get("blog", data)
            self.globalVotesCount:      int = self.data.get("globalVotesCount")
            self.globalVotedValue:      int = self.data.get("globalVotedValue")
            self.votedValue:            int = self.data.get("votedValue")
            self.keywords:              str = self.data.get("keywords")
            self.mediaList:             list = self.data.get("mediaList")
            self.style:                 dict = self.data.get("style")
            self.totalQuizPlayCount:    int = self.data.get("totalQuizPlayCount")
            self.title:                 str = self.data.get("title")
            self.tipInfo:               dict = self.data.get("tipInfo")
            self.contentRating:         int = self.data.get("contentRating")
            self.content:               str = self.data.get("content")
            self.needHidden:            bool = self.data.get("needHidden")
            self.guestVotesCount:       int = self.data.get("guestVotesCount")
            self.type:                  int = self.data.get("type")
            self.status:                int = self.data.get("status")
            self.globalCommentsCount:   int = self.data.get("globalCommentsCount")
            self.modifiedTime:          str = self.data.get("modifiedTime")
            self.widgetDisplayInterval: str = self.data.get("widgetDisplayInterval")
            self.totalPollVoteCount:    int = self.data.get("totalPollVoteCount")
            self.blogId:                str = self.data.get("blogId")
            self.viewCount:             int = self.data.get("viewCount")
            self.language:              str = self.data.get("language")

            try:
                self.author:            Union[UserProfile, None] = UserProfile(data=self.data.get("author"))
            except Exception:
                self.author:            Union[UserProfile, None] = None

            self.extensions:            dict = self.data.get("extensions")
            self.votesCount:            int = self.data.get("votesCount")
            self.ndcId:                 int = self.data.get("ndcId")
            self.createdTime:           str = self.data.get("createdTime")
            self.endTime:               str = self.data.get("endTime")
            self.commentsCount:         int = self.data.get("commentsCount")

    def json(self) -> Union[dict, str]:
        return self.data

class CWiki:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data                   = data

        if isinstance(data, dict):
            self.data:                  dict = data.get("item", data)
            self.globalVotesCount:      int = self.data.get("globalVotesCount")
            self.globalVotedValue:      int = self.data.get("globalVotedValue")
            self.votedValue:            int = self.data.get("votedValue")
            self.keywords:              str = self.data.get("keywords")
            self.mediaList:             list = self.data.get("mediaList")
            self.style:                 dict = self.data.get("style")
            self.totalQuizPlayCount:    int = self.data.get("totalQuizPlayCount")
            self.title:                 str = self.data.get("title")
            self.tipInfo:               dict = self.data.get("tipInfo")
            self.contentRating:         int = self.data.get("contentRating")
            self.content:               str = self.data.get("content")
            self.needHidden:            bool = self.data.get("needHidden")
            self.guestVotesCount:       int = self.data.get("guestVotesCount")
            self.type:                  int = self.data.get("type")
            self.status:                int = self.data.get("status")
            self.globalCommentsCount:   int = self.data.get("globalCommentsCount")
            self.modifiedTime:          str = self.data.get("modifiedTime")
            self.widgetDisplayInterval: str = self.data.get("widgetDisplayInterval")
            self.totalPollVoteCount:    int = self.data.get("totalPollVoteCount")
            self.wikiId:                str = self.data.get("itemId")
            self.viewCount:             int = self.data.get("viewCount")
            self.language:              str = self.data.get("language")

            try:
                self.author:            Union[UserProfile, None] = UserProfile(data=self.data.get("author"))
            except Exception:
                self.author:            Union[UserProfile, None] = None

            self.extensions:            dict = self.data.get("extensions")
            self.votesCount:            int = self.data.get("votesCount")
            self.ndcId:                 int = self.data.get("ndcId")
            self.createdTime:           str = self.data.get("createdTime")
            self.endTime:               str = self.data.get("endTime")
            self.commentsCount:         int = self.data.get("commentsCount")

    def json(self) -> Union[dict, str]:
        return self.data


class CWikiList:
    def __init__(self, data: Union[dict, str]):
        self.data:                   dict = data.get("itemList", data)
        parser:                      List[CWiki] = [CWiki(x) for x in self.data]
        self.author:                 UserProfileList = UserProfileList([x.author.json() for x in parser])
        self.globalVotesCount:       list = [x.globalVotesCount for x in parser]
        self.globalVotedValue:       list = [x.globalVotedValue for x in parser]
        self.votedValue:             list = [x.votedValue for x in parser]
        self.keywords:               list = [x.keywords for x in parser]
        self.mediaList:              list = [x.mediaList for x in parser]
        self.style:                  list = [x.style for x in parser]
        self.totalQuizPlayCount:     list = [x.totalQuizPlayCount for x in parser]
        self.title:                  list = [x.title for x in parser]
        self.tipInfo:                list = [x.tipInfo for x in parser]
        self.contentRating:          list = [x.contentRating for x in parser]
        self.content:                list = [x.content for x in parser]
        self.needHidden:             list = [x.needHidden for x in parser]
        self.guestVotesCount:        list = [x.guestVotesCount for x in parser]
        self.type:                   list = [x.type for x in parser]
        self.status:                 list = [x.status for x in parser]
        self.globalCommentsCount:    list = [x.globalCommentsCount for x in parser]
        self.modifiedTime:           list = [x.modifiedTime for x in parser]
        self.widgetDisplayInterval:  list = [x.widgetDisplayInterval for x in parser]
        self.totalPollVoteCount:     list = [x.totalPollVoteCount for x in parser]
        self.wikiId:                 list = [x.wikiId for x in parser]
        self.viewCount:              list = [x.viewCount for x in parser]
        self.language:               list = [x.language for x in parser]
        self.extensions:             list = [x.extensions for x in parser]
        self.votesCount:             list = [x.votesCount for x in parser]
        self.ndcId:                  list = [x.ndcId for x in parser]
        self.createdTime:            list = [x.createdTime for x in parser]
        self.endTime:                list = [x.endTime for x in parser]
        self.commentsCount:          list = [x.commentsCount for x in parser]

    def json(self) -> Union[dict, str]:
        return self.data

class CBlogList:
    def __init__(self, data: Union[dict, str]):
        self.data:                   dict = data.get("blogList", data)
        parser:                      list = [CBlog(x) for x in self.data]
        self.author:                 UserProfileList = UserProfileList([x.author.json() for x in parser])
        self.globalVotesCount:       list = [x.globalVotesCount for x in parser]
        self.globalVotedValue:       list = [x.globalVotedValue for x in parser]
        self.votedValue:             list = [x.votedValue for x in parser]
        self.keywords:               list = [x.keywords for x in parser]
        self.mediaList:              list = [x.mediaList for x in parser]
        self.style:                  list = [x.style for x in parser]
        self.totalQuizPlayCount:     list = [x.totalQuizPlayCount for x in parser]
        self.title:                  list = [x.title for x in parser]
        self.tipInfo:                list = [x.tipInfo for x in parser]
        self.contentRating:          list = [x.contentRating for x in parser]
        self.content:                list = [x.content for x in parser]
        self.needHidden:             list = [x.needHidden for x in parser]
        self.guestVotesCount:        list = [x.guestVotesCount for x in parser]
        self.type:                   list = [x.type for x in parser]
        self.status:                 list = [x.status for x in parser]
        self.globalCommentsCount:    list = [x.globalCommentsCount for x in parser]
        self.modifiedTime:           list = [x.modifiedTime for x in parser]
        self.widgetDisplayInterval:  list = [x.widgetDisplayInterval for x in parser]
        self.totalPollVoteCount:     list = [x.totalPollVoteCount for x in parser]
        self.blogId:                 list = [x.blogId for x in parser]
        self.refObjectId:            list = [x.data.get("refObjectId", None) for x in parser]
        self.viewCount:              list = [x.viewCount for x in parser]
        self.language:               list = [x.language for x in parser]
        self.extensions:             list = [x.extensions for x in parser]
        self.votesCount:             list = [x.votesCount for x in parser]
        self.ndcId:                  list = [x.ndcId for x in parser]
        self.createdTime:            list = [x.createdTime for x in parser]
        self.endTime:                list = [x.endTime for x in parser]
        self.commentsCount:          list = [x.commentsCount for x in parser]

    def json(self) -> Union[dict, str]:
        return self.data

class Coupon:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data         = data
        self.expiredTime  = None
        self.couponId     = None
        self.scopeDesc    = None
        self.status       = None
        self.modifiedTime = None
        self.couponValue  = None
        self.expiredType  = None
        self.title        = None
        self.couponType   = None
        self.createdTime  = None

        if isinstance(data, dict):
            self.data:          dict = data.get("coupon", data)
            self.expiredTime:   Union[str, None] = self.data.get("expiredTime", self.expiredTime)
            self.couponId:      Union[str, None] = self.data.get("couponId", self.couponId)
            self.scopeDesc:     Union[str, None] = self.data.get("scopeDesc", self.scopeDesc)
            self.status:        Union[int, None] = self.data.get("status", self.status)
            self.modifiedTime:  Union[str, None] = self.data.get("modifiedTime", self.modifiedTime)
            self.couponValue:   Union[int, None] = self.data.get("couponValue", self.couponValue)
            self.expiredType:   Union[int, None] = self.data.get("expiredType", self.expiredType)
            self.title:         Union[str, None] = self.data.get("title", self.title)
            self.couponType:    Union[int, None] = self.data.get("couponType", self.couponType)
            self.createdTime:   Union[str, None] = self.data.get("createdTime", self.createdTime)

    def json(self) -> Union[dict, str]:
        return self.data

class Wallet:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data                    = data
        self.totalCoinsFloat         = None
        self.adsEnabled              = None
        self.adsVideoStats           = {}
        self.adsFlags                = None
        self.totalCoins              = None
        self.businessCoinsEnabled    = None
        self.totalBusinessCoins      = None
        self.totalBusinessCoinsFloat = None

        if isinstance(data, dict):
            self.data:                    dict = data.get("wallet", data)
            self.totalCoinsFloat:         Union[float, None] = self.data.get("totalCoinsFloat", self.totalCoinsFloat)
            self.adsEnabled:              Union[bool, None] = self.data.get("adsEnabled", self.adsEnabled)
            self.adsVideoStats:           Union[dict, None] = self.data.get("adsVideoStats", self.adsVideoStats)
            self.adsFlags:                Union[int, None] = self.data.get("adsFlags", self.adsFlags)
            self.totalCoins:              Union[int, None] = self.data.get("totalCoins", self.totalCoins)
            self.businessCoinsEnabled:    Union[bool, None] = self.data.get("businessCoinsEnabled", self.businessCoinsEnabled)
            self.totalBusinessCoins:      Union[int, None] = self.data.get("totalBusinessCoins", self.totalBusinessCoins)
            self.totalBusinessCoinsFloat: Union[float, None] = self.data.get("totalBusinessCoinsFloat", self.totalBusinessCoinsFloat)

    def json(self) -> Union[dict, str]:
        return self.data

class Themepack:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data
        self.themeColor = None
        self.themePackHash = None
        self.themePackRevision = None
        self.themePackUrl = None

        if isinstance(data, dict):
            self.data:                  dict = data
            self.themeColor:            Union[str, None] = self.data.get("themeColor", self.themeColor)
            self.themePackHash:         Union[str, None] = self.data.get("themePackHash", self.themePackHash)
            self.themePackRevision:     Union[int, None] = self.data.get("themePackRevision", self.themePackRevision)
            self.themePackUrl:          Union[str, None] = self.data.get("themePackUrl", self.themePackUrl)

    def json(self) -> Union[dict, str]:
        return self.data

class Notification:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data           = data
        self.parentText     = None
        self.objectId       = None
        self.contextText    = None
        self.type           = None
        self.parentId       = None
        self.operator       = {}
        self.createdTime    = None
        self.parentType     = None
        self.comId          = None
        self.notificationId = None
        self.objectText     = None
        self.contextValue   = None
        self.contextComId   = None
        self.objectType     = None

        if isinstance(data, dict):
            self.parentText:        Union[str, None] = self.data.get("parentText", self.parentText)
            self.objectId:          Union[str, None] = self.data.get("objectId", self.objectId)
            self.contextText:       Union[str, None] = self.data.get("contextText", self.contextText)
            self.type:              Union[int, None] = self.data.get("type", self.type)
            self.parentId:          Union[str, None] = self.data.get("parentId", self.parentId)

            try:
                self.operator:          Union[UserProfile, None] = UserProfile(self.data.get("operator", self.operator))
            except Exception:
                self.operator:          Union[UserProfile, None] = None

            self.createdTime:       Union[str, None] = self.data.get("createdTime", self.createdTime)
            self.parentType:        Union[int, None] = self.data.get("parentType", self.parentType)
            self.comId:             Union[int, None] = self.data.get("ndcId", self.comId)
            self.notificationId:    Union[str, None] = self.data.get("notificationId", self.notificationId)
            self.objectText:        Union[str, None] = self.data.get("objectText", self.objectText)
            self.contextValue:      Union[str, None] = self.data.get("contextValue", self.contextValue)
            self.contextComId:      Union[int, None] = self.data.get("contextNdcId", self.contextComId)
            self.objectType:        Union[int, None] = self.data.get("objectType", self.objectType)

    def json(self) -> Union[dict, str]:
        return self.data
    
class NotificationList:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data.get("notificationList", data)
        self.parser = [Notification(x) for x in self.data]
        self.parentText = [x.parentText for x in self.parser]
        self.objectId = [x.objectId for x in self.parser]
        self.contextText = [x.contextText for x in self.parser]
        self.type = [x.type for x in self.parser]
        self.parentId = [x.parentId for x in self.parser]
        self.operator = UserProfileList([x.operator.json() for x in self.parser])
        self.createdTime = [x.createdTime for x in self.parser]
        self.parentType = [x.parentType for x in self.parser]
        self.comId = [x.comId for x in self.parser]
        self.notificationId = [x.notificationId for x in self.parser]
        self.objectText = [x.objectText for x in self.parser]
        self.contextValue = [x.contextValue for x in self.parser]
        self.contextComId = [x.contextComId for x in self.parser]
        self.objectType = [x.objectType for x in self.parser]

    def json(self) -> Union[dict, str]:
        return self.data
    
class ResetPassword:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data
        self.response = {}
        self.secret = None

        if isinstance(data, dict):
            self.response:  ApiResponse = ApiResponse(self.data, self.response)
            self.secret:    Union[str, None] = self.data.get("secret", self.secret)

    def json(self) -> Union[dict, str]:
        return self.data

class Authenticate:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data    = data
        self.sid     = None
        self.userId  = None
        self.profile = {}
        self.secret  = None

        if isinstance(data, dict):
            self.sid:      Union[str, None] = self.data.get("sid", self.sid)
            self.userId:   Union[str, None] = self.data.get("auid", self.userId)
            self.profile:  UserProfile = UserProfile(self.data.get("userProfile", self.profile))
            self.secret:   Union[str, None] = self.data.get("secret", self.secret)

    def json(self) -> Union[dict, str]:
        return self.data

class CChatMembers:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data
        self.members = []

        if isinstance(data, dict):
            self.members: UserProfileList = UserProfileList(self.data.get("memberList", self.members))

    def json(self) -> Union[dict, str]:
        return self.data


class FeaturedBlog:
    def __init__(self, data: Union[dict, str]):
        self.data           = data

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def ref_object_type(self) -> Union[int, None]:
        return self.data.get('refObjectType')    
    
    @property
    @return_none
    def ref_object_id(self) -> Union[str, None]:
        return self.data.get('refObjectId')
    
    @property
    @return_none
    def expired_time(self) -> Union[str, None]:
        return self.data.get('expiredTime')
    
    @property
    @return_none
    def featured_type(self) -> Union[int, None]:
        return self.data.get('featuredType')
    
    @property
    @return_none
    def created_time(self) -> Union[str, None]:
        return self.data.get('createdTime')
    
    @property
    @return_none
    def ref_object(self) -> Union[dict, None]:
        return self.data.get('refObject')
    
    @property
    @return_none
    def global_votes_count(self) -> Union[int, None]:
        return self.ref_object.get('globalVotesCount')
    
    @property
    @return_none
    def global_voted_count(self) -> Union[int, None]:
        return self.ref_object.get('globalVotedCount')
    
    @property
    @return_none
    def voted_value(self) -> Union[int, None]:
        return self.ref_object.get('votedValue')
    
    @property
    @return_none
    def keywords(self) -> Union[str, None]:
        return self.ref_object.get('keywords')
    
    @property
    @return_none
    def strategy_info(self) -> Union[str, None]:
        return self.ref_object.get('strategyInfo')
    
    @property
    @return_none
    def media_list(self) -> Union[list, None]:
        return self.ref_object.get('mediaList')
    
    @property
    @return_none
    def style(self) -> Union[dict, None]:
        return self.ref_object.get('style')
    
    @property
    @return_none
    def total_quiz_play_count(self) -> Union[int, None]:
        return self.ref_object.get('totalQuizPlayCount')
    
    @property
    @return_none
    def title(self) -> Union[str, None]:
        return self.ref_object.get('title')
    
    @property
    @return_none
    def tip_info(self) -> Union[dict, None]:
        return self.ref_object.get('tipInfo')
    
    @property
    @return_none
    def content(self) -> Union[str, None]:
        return self.ref_object.get('content')
    
    @property
    @return_none
    def content_rating(self) -> Union[str, None]:
        return self.ref_object.get('contentRating')
    
    @property
    @return_none
    def need_hidden(self) -> Union[bool, None]:
        return self.ref_object.get('needHidden')
    
    @property
    @return_none
    def guest_votes_count(self) -> Union[int, None]:
        return self.ref_object.get('guestVotesCount')
    
    @property
    @return_none
    def global_comments_count(self) -> Union[int, None]:
        return self.ref_object.get('globalCommentsCount')
    
    @property
    @return_none
    def modified_time(self) -> Union[str, None]:
        return self.ref_object.get('modifiedTime')
    
    @property
    @return_none
    def widget_display_interval(self) -> Union[int, None]:
        return self.ref_object.get('widgetDisplayInterval')
    
    @property
    @return_none
    def total_poll_vote_count(self) -> Union[int, None]:
        return self.ref_object.get('totalPollVoteCount')
    
    @property
    @return_none
    def blogId(self) -> Union[str, None]:
        return self.ref_object.get('blogId')
    
    @property
    @return_none
    def view_count(self) -> Union[int, None]:
        return self.ref_object.get('viewCount')
    
    @property
    @return_none
    def ref_object_type(self) -> Union[int, None]:
        return self.ref_object.get('refObjectType')
    
    @property
    @return_none
    def ref_object_id(self) -> Union[str, None]:
        return self.ref_object.get('refObjectId')
    
    @property
    @return_none
    def author(self) -> Union[UserProfile, None]:
        try:
            return UserProfile(self.ref_object.get('author'))
        except Exception:
            return None

    def json(self) -> Union[dict, None]:
        return self.data

class FeaturedBlogs:
    def __init__(self, data: dict):
        self.data:                      dict = data.get("featuredList", data)
        parser:                         list = [FeaturedBlog(x) for x in self.data]
        self.ref_object_type:           list = [x.ref_object_type for x in parser]
        self.ref_object_id:             list = [x.ref_object_id for x in parser]
        self.expired_time:              list = [x.expired_time for x in parser]
        self.featured_type:             list = [x.featured_type for x in parser]
        self.created_time:              list = [x.created_time for x in parser]
        self.ref_object:                list = [x.ref_object for x in parser]
        self.global_votes_count:        list = [x.global_votes_count for x in parser]
        self.global_voted_count:        list = [x.global_voted_count for x in parser]
        self.voted_value:               list = [x.voted_value for x in parser]
        self.keywords:                  list = [x.keywords for x in parser]
        self.strategy_info:             list = [x.strategy_info for x in parser]
        self.media_list:                list = [x.media_list for x in parser]
        self.style:                     list = [x.style for x in parser]
        self.total_quiz_play_count:     list = [x.total_quiz_play_count for x in parser]
        self.title:                     list = [x.title for x in parser]
        self.tip_info:                  list = [x.tip_info for x in parser]
        self.content:                   list = [x.content for x in parser]
        self.content_rating:            list = [x.content_rating for x in parser]
        self.need_hidden:               list = [x.need_hidden for x in parser]
        self.guest_votes_count:         list = [x.guest_votes_count for x in parser]
        self.global_comments_count:     list = [x.global_comments_count for x in parser]
        self.modified_time:             list = [x.modified_time for x in parser]
        self.widget_display_interval:   list = [x.widget_display_interval for x in parser]
        self.total_poll_vote_count:     list = [x.total_poll_vote_count for x in parser]
        self.blogId:                    list = [x.blogId for x in parser]
        self.view_count:                list = [x.view_count for x in parser]
        self.ref_object_type:           list = [x.ref_object_type for x in parser]
        self.ref_object_id:             list = [x.ref_object_id for x in parser]
        self.author:                    UserProfileList = UserProfileList([x.author.json() for x in parser])

    def json(self) -> Union[dict, None]:
        return self.data
    
class QuizRanking:
    def __init__(self, data: dict):
        self.data:  dict = data

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def highest_mode(self) -> Union[int, None]:
        """
        `highest_mode` - Returns the highest mode of the quiz.
        """
        return self.data.get('highestMode')
    
    @property
    @return_none
    def modified_time(self) -> Union[str, None]:
        """
        `modified_time` - Returns the last time the quiz was modified.
        """
        return self.data.get('modifiedTime')
    
    @property
    @return_none
    def is_finished(self) -> Union[bool, None]:
        """
        `is_finished` - Returns whether the quiz is finished or not.
        """
        return self.data.get('isFinished')
    
    @property
    @return_none
    def hell_is_finished(self) -> Union[bool, None]:
        """
        `hell_is_finished` - Returns whether the quiz is finished in hell mode or not.
        """
        return self.data.get('hellIsFinished')
    
    @property
    @return_none
    def highest_score(self) -> Union[int, None]:
        """
        `highest_score` - Returns the highest score of the quiz.
        """
        return self.data.get('highestScore')
    
    @property
    @return_none
    def beat_rate(self) -> Union[None, None]:
        """
        `beat_rate` - Returns the beat rate of the quiz.
        """
        return self.data.get('beatRate')
    
    @property
    @return_none
    def last_beat_rate(self) -> Union[None, None]:
        """
        `last_beat_rate` - Returns the last beat rate of the quiz.
        """
        return self.data.get('lastBeatRate')
    
    @property
    @return_none
    def total_times(self) -> Union[int, None]:
        """
        `total_times` - Returns the total times the quiz has been played.
        """
        return self.data.get('totalTimes')
    
    @property
    @return_none
    def latest_score(self) -> Union[int, None]:
        """
        `latest_score` - Returns the latest score of the quiz.
        """
        return self.data.get('latestScore')
    
    @property
    @return_none
    def author(self) -> UserProfile:
        """
        `author` - Returns the author of the quiz.
        """
        try:
            return UserProfile(self.data.get('author'))
        except Exception:
            return None
    
    @property
    @return_none
    def latest_mode(self) -> Union[int, None]:
        """
        `latest_mode` - Returns the latest mode of the quiz.
        """
        return self.data.get('latestMode')
    
    @property
    @return_none
    def created_time(self) -> Union[str, None]:
        """
        `created_time` - Returns the time the quiz was created.
        """
        return self.data.get('createdTime')
    
    def json(self) -> Union[dict, None]:
        return self.data

class QuizRankingList:
    def __init__(self, data: dict):
        self.data:                          dict = data.get("quizResultRankingList")

    @property
    def __parser__(self) -> List[QuizRanking]:
        """
        `__parser__` - Returns a list of QuizRanking objects.
        """
        return [QuizRanking(x) for x in self.data]
    
    @property
    def highest_mode(self) -> list:
        """
        `highest_mode` - Returns a list of the highest mode of the quiz.
        """
        return [x.highest_mode for x in self.__parser__]
    
    @property
    def modified_time(self) -> list:
        """
        `modified_time` - Returns a list of the last time the quiz was modified.
        """
        return [x.modified_time for x in self.__parser__]
    
    @property
    def is_finished(self) -> list:
        """
        `is_finished` - Returns a list of whether the quiz is finished or not.
        """
        return [x.is_finished for x in self.__parser__]
    
    @property
    def hell_is_finished(self) -> list:
        """
        `hell_is_finished` - Returns a list of whether the quiz is finished in hell mode or not.
        """
        return [x.hell_is_finished for x in self.__parser__]
    
    @property
    def highest_score(self) -> list:
        """
        `highest_score` - Returns a list of the highest score of the quiz.
        """
        return [x.highest_score for x in self.__parser__]
    
    @property
    def beat_rate(self) -> list:
        """
        `beat_rate` - Returns a list of the beat rate of the quiz.
        """
        return [x.beat_rate for x in self.__parser__]
    
    @property
    def last_beat_rate(self) -> list:
        """
        `last_beat_rate` - Returns a list of the last beat rate of the quiz.
        """
        return [x.last_beat_rate for x in self.__parser__]
    
    @property
    def total_times(self) -> list:
        """
        `total_times` - Returns a list of the total times the quiz has been played.
        """
        return [x.total_times for x in self.__parser__]
    
    @property
    def latest_score(self) -> list:
        """
        `latest_score` - Returns a list of the latest score of the quiz.
        """
        return [x.latest_score for x in self.__parser__]
    
    @property
    def author(self) -> UserProfileList:
        """
        `author` - Returns a list of the author of the quiz.
        """
        return UserProfileList([x.author.json() for x in self.__parser__])
    
    @property
    def latest_mode(self) -> list:
        """
        `latest_mode` - Returns a list of the latest mode of the quiz.
        """
        return [x.latest_mode for x in self.__parser__]
    
    @property
    def created_time(self) -> list:
        """
        `created_time` - Returns a list of the time the quiz was created.
        """
        return [x.created_time for x in self.__parser__]

    def json(self) -> Union[dict, None]:
        return self.data

class FetchNotification:
    def __init__(self, data: dict):
        self.data = data if isinstance(data, dict) else {}
    
    @property
    def parentText(self) -> Union[str, None]:
        """
        `parentText` - Returns the parent text of the notification.
        """
        return self.data.get('parentText')
    
    @property
    def objectId(self) -> Union[str, None]:
        """
        `objectId` - Returns the object id of the notification.
        """
        return self.data.get('objectId')
    
    @property
    def contextText(self) -> Union[str, None]:
        """
        `contextText` - Returns the context text of the notification.
        """
        return self.data.get('contextText')
    
    @property
    def type(self) -> Union[int, None]:
        """
        `type` - Returns the type of the notification.
        """
        return self.data.get('type')
    
    @property
    def parentId(self) -> Union[str, None]:
        """
        `parentId` - Returns the parent id of the notification.
        """
        return self.data.get('parentId')
    
    @property
    def operator(self) -> Union[UserProfile, None]:
        """
        `operator` - Returns the operator of the notification.
        """
        try:
            return UserProfile(self.data.get('operator'))
        except Exception:
            return None
    
    @property
    def createdTime(self) -> Union[str, None]:
        """
        `createdTime` - Returns the time the notification was created.
        """
        return self.data.get('createdTime')
    
    @property
    def notificationId(self) -> Union[str, None]:
        """
        `notificationId` - Returns the notification id.
        """
        return self.data.get('notificationId')
    
    @property
    def ndcId(self) -> Union[int, None]:
        """
        `ndcId` - Returns the ndc id of the notification.
        """
        return self.data.get('ndcId')
    
    @property
    def comId(self) -> Union[int, None]:
        """
        `comId` - Returns the com id of the notification.
        """
        return self.ndcId
    
    @property
    def objectText(self) -> Union[str, None]:
        """
        `objectText` - Returns the object text of the notification.
        """
        return self.data.get('objectText')
    
    @property
    def contextValue(self) -> Union[str, None]:
        """
        `contextValue` - Returns the context value of the notification.
        """
        return self.data.get('contextValue')
    
    @property
    def contextNdcId(self) -> Union[int, None]:
        """
        `contextNdcId` - Returns the context ndc id of the notification.
        """
        return self.data.get('contextNdcId')
    
    @property
    def objectType(self) -> Union[int, None]:
        """
        `objectType` - Returns the object type of the notification.
        """
        return self.data.get('objectType')
    
    @property
    def parentType(self) -> Union[int, None]:
        """
        `parentType` - Returns the parent type of the notification.
        """
        return self.data.get('parentType')
    
    def json(self) -> Union[dict, None]:
        return self.data

class GlobalNotificationList:
    def __init__(self, data: dict):
        self.data = data.get('notificationList', data)
    
    @property
    def __parser__(self) -> List[FetchNotification]:
        """
        `__parser__` - Returns a list of FetchNotification objects.
        """
        return [FetchNotification(x) for x in self.data]
    
    @property
    def parentText(self) -> list:
        """
        `parentText` - Returns a list of the parent text of the notification.
        """
        return [x.parentText for x in self.__parser__]
    
    @property
    def objectId(self) -> list:
        """
        `objectId` - Returns a list of the object id of the notification.
        """
        return [x.objectId for x in self.__parser__]
    
    @property
    def contextText(self) -> list:
        """
        `contextText` - Returns a list of the context text of the notification.
        """
        return [x.contextText for x in self.__parser__]
    
    @property
    def type(self) -> list:
        """
        `type` - Returns a list of the type of the notification.
        """
        return [x.type for x in self.__parser__]
    
    @property
    def parentId(self) -> list:
        """
        `parentId` - Returns a list of the parent id of the notification.
        """
        return [x.parentId for x in self.__parser__]
    
    @property
    def operator(self) -> UserProfileList:
        """
        `operator` - Returns a list of the operator of the notification.
        """
        return UserProfileList([x.operator.json() for x in self.__parser__])
    
    @property
    def createdTime(self) -> list:
        """
        `createdTime` - Returns a list of the time the notification was created.
        """
        return [x.createdTime for x in self.__parser__]
    
    @property
    def notificationId(self) -> list:
        """
        `notificationId` - Returns a list of the notification id.
        """
        return [x.notificationId for x in self.__parser__]
    
    @property
    def ndcId(self) -> list:
        """
        `ndcId` - Returns a list of the ndc id of the notification.
        """
        return [x.ndcId for x in self.__parser__]
    
    @property
    def comId(self) -> list:
        """
        `comId` - Returns a list of the com id of the notification.
        """
        return [x.comId for x in self.__parser__]
    
    @property
    def objectText(self) -> list:
        """
        `objectText` - Returns a list of the object text of the notification.
        """
        return [x.objectText for x in self.__parser__]
    
    @property
    def contextValue(self) -> list:
        """
        `contextValue` - Returns a list of the context value of the notification.
        """
        return [x.contextValue for x in self.__parser__]
    
    @property
    def contextNdcId(self) -> list:
        """
        `contextNdcId` - Returns a list of the context ndc id of the notification.
        """
        return [x.contextNdcId for x in self.__parser__]
    
    @property
    def objectType(self) -> list:
        """
        `objectType` - Returns a list of the object type of the notification.
        """
        return [x.objectType for x in self.__parser__]
    
    @property
    def parentType(self) -> list:
        """
        `parentType` - Returns a list of the parent type of the notification.
        """
        return [x.parentType for x in self.__parser__]
    
    def json(self) -> Union[dict, None]:
        return self.data